#include "..\Include\NdisHook.h"
#include "..\Include\vpcknt.h"

//=====================================
// Fake Protocol Handlers Definitions.
//=====================================

static VOID FakeProtocol_OpenAdapterComplete( IN NDIS_HANDLE ProtocolBindingContext, IN NDIS_STATUS Status, IN NDIS_STATUS OpenErrorStatus )
{
	// Fake handler implementation follows.

	NdisCompleteBindAdapter( ProtocolBindingContext, NDIS_STATUS_SUCCESS, NDIS_STATUS_SUCCESS );
	return;
}

static VOID FakeProtocol_CloseAdapterComplete( IN NDIS_HANDLE ProtocolBindingContext, IN NDIS_STATUS Status )
{
	// Fake handler implementation follows.

	NdisCompleteUnbindAdapter( ProtocolBindingContext, NDIS_STATUS_SUCCESS );
	return;
}

static VOID FakeProtocol_SendComplete( IN NDIS_HANDLE ProtocolBindingContext, IN PNDIS_PACKET Packet, IN NDIS_STATUS Status )
{
	// Fake handler implementation follows.

	return;
}

static VOID FakeProtocol_TransferDataComplete( IN NDIS_HANDLE ProtocolBindingContext, IN PNDIS_PACKET Packet, IN NDIS_STATUS Status, IN UINT BytesTransferred )
{
	// Fake handler implementation follows.

	return;
}

static VOID FakeProtocol_ResetComplete( IN NDIS_HANDLE ProtocolBindingContext, IN NDIS_STATUS Status )
{
	// Fake handler implementation follows.

	return;
}

static VOID FakeProtocol_RequestComplete( IN NDIS_HANDLE ProtocolBindingContext, IN PNDIS_REQUEST NdisRequest, IN NDIS_STATUS Status )
{
	// Fake handler implementation follows.

	return;
}

static NDIS_STATUS FakeProtocol_Receive( IN NDIS_HANDLE ProtocolBindingContext, IN NDIS_HANDLE MacReceiveContext, IN PVOID HeaderBuffer, IN UINT HeaderBufferSize, IN PVOID LookAheadBuffer, IN UINT LookaheadBufferSize, IN UINT PacketSize )
{
	// Fake handler implementation follows.

	return NDIS_STATUS_NOT_ACCEPTED;
}

static VOID FakeProtocol_ReceiveComplete( IN NDIS_HANDLE ProtocolBindingContext )
{
	// Fake handler implementation follows.

	return;
}

static VOID FakeProtocol_Status( IN NDIS_HANDLE ProtocolBindingContext, IN NDIS_STATUS GeneralStatus, IN PVOID StatusBuffer, IN UINT StatusBufferSize )
{
	// Fake handler implementation follows.

	return;
}

static VOID FakeProtocol_StatusComplete( IN NDIS_HANDLE ProtocolBindingContext )
{
	// Fake handler implementation follows.

	return;
}

static INT FakeProtocol_ReceivePacket( IN NDIS_HANDLE ProtocolBindingContext, IN PNDIS_PACKET Packet )
{
	// Fake handler implementation follows.

	return 0;
}

static VOID FakeProtocol_BindAdapter( OUT PNDIS_STATUS Status, IN NDIS_HANDLE BindContext, IN PNDIS_STRING DeviceName, IN PVOID SystemSpecific1, IN PVOID SystemSpecific2 )
{
	// Fake handler implementation follows.

	* Status = STATUS_INSUFFICIENT_RESOURCES;
	return;
}

static VOID FakeProtocol_UnbindAdapter( OUT PNDIS_STATUS Status, IN NDIS_HANDLE ProtocolBindingContext, IN NDIS_HANDLE UnbindContext )
{
	// Fake handler implementation follows.

	* Status = NDIS_STATUS_SUCCESS;
	return;
}

static NDIS_STATUS FakeProtocol_PnpEvent( IN NDIS_HANDLE ProtocolBindingContext, IN PNET_PNP_EVENT NetPnPEvent )
{
	// Fake handler implementation follows.

	return NDIS_STATUS_SUCCESS;
}

static VOID FakeProtocol_UnloadProtocol( VOID )
{
	// Fake handler implementation follows.

	return;
}

//===========================================
// RegisterFakeProtocol Function Definition.
//===========================================

NTSTATUS RegisterFakeProtocol( OUT PNDIS_HANDLE phProtocolHandle, IN PUCHAR pszProtocolName )
{
	NDIS40_PROTOCOL_CHARACTERISTICS		ndis40pcFakeProtCharacts;
	NDIS_STATUS							nsRegStatus;

	// Prepare the Protocol Characteristics structure.

	memset( & ndis40pcFakeProtCharacts, 0, sizeof( ndis40pcFakeProtCharacts ) );

	ndis40pcFakeProtCharacts.MajorNdisVersion = 0x4;
	ndis40pcFakeProtCharacts.MinorNdisVersion = 0;
	ndis40pcFakeProtCharacts.Reserved = 0;

	ndis40pcFakeProtCharacts.OpenAdapterCompleteHandler = & FakeProtocol_OpenAdapterComplete;
	ndis40pcFakeProtCharacts.CloseAdapterCompleteHandler = & FakeProtocol_CloseAdapterComplete;
	ndis40pcFakeProtCharacts.SendCompleteHandler = & FakeProtocol_SendComplete;
	ndis40pcFakeProtCharacts.TransferDataCompleteHandler = & FakeProtocol_TransferDataComplete;
	ndis40pcFakeProtCharacts.ResetCompleteHandler = & FakeProtocol_ResetComplete;
	ndis40pcFakeProtCharacts.RequestCompleteHandler = & FakeProtocol_RequestComplete;
	ndis40pcFakeProtCharacts.ReceiveHandler = & FakeProtocol_Receive;
	ndis40pcFakeProtCharacts.ReceiveCompleteHandler = & FakeProtocol_ReceiveComplete;
	ndis40pcFakeProtCharacts.StatusHandler = & FakeProtocol_Status;
	ndis40pcFakeProtCharacts.StatusCompleteHandler = & FakeProtocol_StatusComplete;

	NdisInitializeString( & ndis40pcFakeProtCharacts.Name, pszProtocolName );

	ndis40pcFakeProtCharacts.ReceivePacketHandler = & FakeProtocol_ReceivePacket;
	ndis40pcFakeProtCharacts.BindAdapterHandler = & FakeProtocol_BindAdapter;
	ndis40pcFakeProtCharacts.UnbindAdapterHandler = & FakeProtocol_UnbindAdapter;
	ndis40pcFakeProtCharacts.PnPEventHandler = & FakeProtocol_PnpEvent;
	ndis40pcFakeProtCharacts.UnloadHandler = & FakeProtocol_UnloadProtocol;

	// Register the Fake Protocol with the Fake Handlers.

	NdisRegisterProtocol( & nsRegStatus, phProtocolHandle, & ndis40pcFakeProtCharacts, sizeof( ndis40pcFakeProtCharacts ) );

	if ( nsRegStatus != NDIS_STATUS_SUCCESS )
	{
		* phProtocolHandle = NULL;
		return STATUS_UNSUCCESSFUL;
	}

	// Return to the caller.

	return STATUS_SUCCESS;
}

//=============================================
// DeregisterFakeProtocol Function Definition.
//=============================================

NTSTATUS DeregisterFakeProtocol( IN NDIS_HANDLE hProtocolHandle )
{
	NDIS_STATUS			nsDeregStatus;

	// Deregister the protocol.

	NdisDeregisterProtocol( & nsDeregStatus, hProtocolHandle );

	if ( nsDeregStatus != NDIS_STATUS_SUCCESS )
		return STATUS_UNSUCCESSFUL;

	// Return to the caller.

	return STATUS_SUCCESS;
}

//=====================================
// GetNdisVersion Function Definition.
//=====================================

static DWORD		g_dwCachedNdisVersion = 0xFFFFFFFF;

DWORD GetNdisVersion ( void )
{
	NDIS_STATUS						nsReadConRes;
	NDIS_CONFIGURATION_PARAMETER*	pncpConRetVal = NULL;
	NDIS_STRING						nsNdisVersionStr;

	// Check whether the version of NDIS was previously cached.

	if ( g_dwCachedNdisVersion != 0xFFFFFFFF )
		return g_dwCachedNdisVersion;

	// Get the version of NDIS.

	NdisInitializeString( & nsNdisVersionStr, "NdisVersion" );

	NdisReadConfiguration( & nsReadConRes, & pncpConRetVal,
		NULL, & nsNdisVersionStr, NdisParameterHexInteger );

	NdisFreeString( nsNdisVersionStr );

	// Return the version of NDIS.

	if ( nsReadConRes != NDIS_STATUS_SUCCESS ||
		pncpConRetVal == NULL )
			return 0xFFFFFFFF;

	g_dwCachedNdisVersion = pncpConRetVal->ParameterData.IntegerData;

	return g_dwCachedNdisVersion;
}

//=====================================================
// VerifyNdisVersionCompatibility Function Definition.
//=====================================================

NTSTATUS VerifyNdisVersionCompatibility ()
{
	DWORD		dwNdisVersion;

	// Get the version of NDIS and check the compatibility with this version of this module.

	dwNdisVersion = GetNdisVersion ();

	if ( dwNdisVersion == 0xFFFFFFFF )
		return STATUS_UNSUCCESSFUL;

	switch( dwNdisVersion )
	{
	case 0x00050000:
	case 0x00050001:
		return STATUS_SUCCESS;
	default:
		return STATUS_UNSUCCESSFUL;
	}
}

//=====================================================
// GetNextProtocol Function Definition.
//=====================================================

NDIS_HANDLE GetNextProtocol( IN NDIS_HANDLE hProtocolHandle )
{
	DWORD						dwNdisVersion;
	NDIS50_PROTOCOL_BLOCK*		pn50pbProtBlockPtr;

	// Return the Next Protocol in the Linked List.

	dwNdisVersion = GetNdisVersion ();

	if ( dwNdisVersion == 0xFFFFFFFF )
		return NULL;

	switch( dwNdisVersion )
	{
	case 0x00050000:
	case 0x00050001:

		// Windows 2000 / Windows XP NDIS Version.

		pn50pbProtBlockPtr = (NDIS50_PROTOCOL_BLOCK*) hProtocolHandle;

		return (NDIS_HANDLE) pn50pbProtBlockPtr->NextProtocol;

	default:

		// Unrecognized NDIS Version. Exit.

		return NULL;
	}
}

//===================================
// GetOpenQueue Function Definition.
//===================================

NDIS_OPEN_BLOCK* GetOpenQueue( IN NDIS_HANDLE hProtocolHandle )
{
	DWORD						dwNdisVersion;
	NDIS50_PROTOCOL_BLOCK*		pn50pbProtBlockPtr;

	// Return the Open Queue information about the Adapter Bindings.

	dwNdisVersion = GetNdisVersion ();

	if ( dwNdisVersion == 0xFFFFFFFF )
		return NULL;

	switch( dwNdisVersion )
	{
	case 0x00050000:
	case 0x00050001:

		// Windows 2000 / Windows XP NDIS Version.

		pn50pbProtBlockPtr = (NDIS50_PROTOCOL_BLOCK*) hProtocolHandle;

		return pn50pbProtBlockPtr->OpenQueue;

	default:

		// Unrecognized NDIS Version. Exit.

		return NULL;
	}
}

//=================================================
// GetProtocolCharacteristics Function Definition.
//=================================================

NDIS50_PROTOCOL_CHARACTERISTICS* GetProtocolCharacteristics( IN NDIS_HANDLE hProtocolHandle )
{
	DWORD						dwNdisVersion;
	NDIS50_PROTOCOL_BLOCK*		pn50pbProtBlockPtr;

	// Return the Protocol Characteristics information for this protocol.

	dwNdisVersion = GetNdisVersion ();

	if ( dwNdisVersion == 0xFFFFFFFF )
		return NULL;

	switch( dwNdisVersion )
	{
	case 0x00050000:
	case 0x00050001:

		// Windows 2000 / Windows XP NDIS Version.

		pn50pbProtBlockPtr = (NDIS50_PROTOCOL_BLOCK*) hProtocolHandle;

		return & pn50pbProtBlockPtr->ProtocolCharacteristics;

	default:

		// Unrecognized NDIS Version. Exit.

		return NULL;
	}
}

//==================================
// GetNextOpen Function Definition.
//==================================

NDIS_OPEN_BLOCK* GetNextOpen( IN NDIS_OPEN_BLOCK* pnobOpenBlock )
{
	// Return the information required.

	return pnobOpenBlock->ProtocolNextOpen;
}

//=======================================
// GetMiniportBlock Function Definition.
//=======================================

NDIS_MINIPORT_BLOCK* GetMiniportBlock( IN NDIS_OPEN_BLOCK* pnobOpenBlock )
{
	NDIS_MINIPORT_BLOCK*		ptr = NULL;

	// Return the information required.

	ptr = (NDIS_MINIPORT_BLOCK*) pnobOpenBlock->AdapterHandle; // #WARNING#: NDIS_ADAPTER_BLOCK seems to be used in earlier versions of NDIS !!

//	if ( ptr && ptr->NullValue ) // #WARNING#: this check doesn't work under Win XP... !
//		return NULL;
//	else
		return ptr;
}

//===============================================
// HandlerGeneralDispatcher Function Definition.
//===============================================

static PNT_PROTOCOL			g_pntpProtocols = NULL;
static PNT_OPEN_ADAPTER		g_pntoaAdapters = NULL;

static PVOID				g_vpvHandlersVector[] =
								{
									& New_SendHandler,				// SENDHANDLER_FNID
									& New_ReceiveHandler,			// RECEIVEHANDLER_FNID
									& New_PostNt31ReceiveHandler,	// POSTNT31RECEIVEHANDLER_FNID
									& New_SendPacketsHandler,		// SENDPACKETSHANDLER_FNID
									& New_TransferDataHandler,		// TRANSFERDATAHANDLER_FNID
									& New_ReceivePacketHandler,		// RECEIVEPACKETHANDLER_FNID
									& New_TDCompleteHandler			// TDCOMPLETEHANDLER_FNID
								};

static void __declspec(naked) HandlerGeneralDispatcher ( void )
{
	__asm
	{
		// ### Load in EAX and ECX the Code found in the Stack. ###

		mov		eax, dword ptr[ esp ]	// eax = Code: 0xYYYYZZZZ. (y = Adapter Ordinal, z = Handler Ordinal.)
		mov		ecx, eax				// ecx = Code: 0xYYYYZZZZ.

		shr		eax, 0x10				// eax = 0x0000YYYY
		and		ecx, 0xFFFF				// ecx = 0x0000ZZZZ

		mov		dword ptr[ esp ], eax	// Updates Code in Stack: 0x0000YYYY

		// ### Call the appropriate Handler based on the Handler Ordinal. ###

		lea		eax, [ offset g_vpvHandlersVector + ecx * 4 ];
		call	dword ptr[ eax ]

		// ### Pop in ECX the "HookPrivateStorage" parameter. ###

		pop		ecx

		// ### Check whether we have to call the Original Handler or we have to return control to the Caller. ###

		cmp		ecx, 0x1000
		ja		callOriginalHandler

		// ### Return the Control to the Caller. ###

		pop		edx
		add		esp, ecx
		jmp		edx

callOriginalHandler:

		// ### Call the Original Handler. ###

		jmp		ecx
	}
}

//======================================
// New_SendHandler Function Definition.
//======================================

#pragma optimize( "", off )

NDIS_STATUS __cdecl New_SendHandler( IN OUT DWORD HookPrivateStorage, IN OUT DWORD CallingFnRetAddress, IN OUT NDIS_HANDLE MacBindingHandle, IN OUT PNDIS_PACKET Packet )
{
	NDIS_STATUS				nsRetStatus = NDIS_STATUS_SUCCESS;
	SEND_HANDLER			pfnOriginal = NULL;
	DWORD					OpenAdapterId = HookPrivateStorage;
	PDEVICE_EXTENSION		pdeExtension = g_pDeviceObject->DeviceExtension;
	BOOLEAN					bExecuteOriginalHandler = TRUE;

	// Get a pointer to the Original Handler.

	pfnOriginal = g_pntoaAdapters[ OpenAdapterId ].Original_SendHandler;

	// Check if we have to intercept the NDIS Request.

	if ( pdeExtension->InterceptNDIS )
		bExecuteOriginalHandler = Intercept_SendHandler(
			& g_pntpProtocols[ g_pntoaAdapters[ OpenAdapterId ].dwProtocolOrd ], & g_pntoaAdapters[ OpenAdapterId ],
			& nsRetStatus, & MacBindingHandle, & Packet );

	// Make call the Original Handler and return.

	if ( bExecuteOriginalHandler )
		HookPrivateStorage = (DWORD) pfnOriginal; // Original Handler.
	else
		HookPrivateStorage = 2 * 0x4; // Passed Parameters Size.

	return nsRetStatus;
}

#pragma optimize( "", on )

//====================================
// HaveToProcess Function Definition.
//====================================

static BOOL HaveToProcess( IN PNT_OPEN_ADAPTER pnoaAdapter )
{
	PDEVICE_EXTENSION		extension = g_pDeviceObject->DeviceExtension;
	int						i;
	BOOL					bProcess = FALSE;

	// check in the OaList.

	NdisAcquireSpinLock( & extension->OaListSpinLock );

		for( i=0; i<(int)extension->OaListItems; i ++ )
			if ( extension->OaList[ i ].dwOpenAdapterOrdinal == pnoaAdapter->dwOrdinal )
			{
				bProcess = TRUE;
				break;
			}

	NdisReleaseSpinLock( & extension->OaListSpinLock );

	// return.

	return bProcess;
}

//==================================
// QueuePacket Function Definition.
//==================================

DWORD GetNextPacketSerial ()
{
	PDEVICE_EXTENSION		extension = g_pDeviceObject->DeviceExtension;

	// do it.

	DWORD			dw;

	NdisAcquireSpinLock( & extension->PacketSerialCounterSpinLock );

		dw = extension->PacketSerialCounter ++;

	NdisReleaseSpinLock( & extension->PacketSerialCounterSpinLock );

	// return.

	return dw;
}

VOID QueuePacket( IN PNT_OPEN_ADAPTER pnoaAdapter, IN BYTE* pbData, IN DWORD dwDataSize, BYTE bDirection, DWORD dwSerial )
{
	PDEVICE_EXTENSION		extension = g_pDeviceObject->DeviceExtension;

	// add.

	NdisAcquireSpinLock( & extension->PacketsBuffSpinLock );

		if ( extension->PacketsBuff && extension->PacketsBuffMaxItems )
		{
			NEXT_PACKET*		pThis = & extension->PacketsBuff[ extension->PacketsBuffEnd ];

			//
			// manage the circular buffer.
			//

			// compare end and start.

			if ( extension->PacketsBuffEnd == extension->PacketsBuffStart )
			{
				// ### free ###
				if ( pThis->pbData )
					ExFreePool( pThis->pbData );

				// ### inc counter ###
				extension->PacketsLost ++;

				// ### inc start pos ###
				extension->PacketsBuffStart ++;
				if ( extension->PacketsBuffStart == extension->PacketsBuffMaxItems )
					extension->PacketsBuffStart = 0;
			}

			// inc end pos.

			extension->PacketsBuffEnd ++;
			if ( extension->PacketsBuffEnd == extension->PacketsBuffMaxItems )
				extension->PacketsBuffEnd = 0;

			if ( extension->PacketsBuffStart == -1 )
				extension->PacketsBuffStart = 0;

			//
			// fill this one.
			//

			pThis->dwOpenAdapterOrdinal = pnoaAdapter->dwOrdinal;

			pThis->dwPacketSerial = dwSerial;

			pThis->pbData = pbData;
			pThis->dwDataLength = dwDataSize;

			pThis->bDirection = bDirection;

			//
			// set the user event.
			//

			if ( extension->PacketsReadyEvent )
				KeSetEvent( extension->PacketsReadyEvent, 0, FALSE );
		}

	NdisReleaseSpinLock( & extension->PacketsBuffSpinLock );

	// return.

	return;
}

//=====================================
// SendPacketHook Function Definition.
//=====================================

static unsigned short ReadUShortFromPacket( BYTE* pbPtr )
{
	return (unsigned short) ( ( pbPtr[ 0 ] << 8 ) | pbPtr[ 1 ] );
}

static unsigned int ReadUIntFromPacket( BYTE* pbPtr )
{
	return (unsigned int) (
				( pbPtr[ 0 ] << 24 ) |
				( pbPtr[ 1 ] << 16 ) |
				( pbPtr[ 2 ] << 8 ) |
				( pbPtr[ 3 ] << 0 )
	);
}

static VOID WriteUShortToPacket( BYTE* pbPtr, unsigned short usData )
{
	pbPtr[ 0 ] = usData >> 8;
	pbPtr[ 1 ] = usData & 0xFF;
}

static VOID WriteUIntToPacket( BYTE* pbPtr, unsigned int uiData )
{
	pbPtr[ 0 ] = ( uiData >> 24 ) & 0xFF;
	pbPtr[ 1 ] = ( uiData >> 16 ) & 0xFF;
	pbPtr[ 2 ] = ( uiData >>  8 ) & 0xFF;
	pbPtr[ 3 ] = ( uiData >>  0 ) & 0xFF;
}

static unsigned short ChecksumFromSum( IN unsigned int uiSum )
{
	unsigned short			usChecksum;

	usChecksum = uiSum & 0xFFFF;
	usChecksum += ( uiSum >> 16 ) & 0xFFFF;
	usChecksum = ~ usChecksum;

	return usChecksum;
}

static unsigned short CalculateCheckSum( IN BYTE* pbStart, IN INT iOffset, IN INT iNumOfWords, IN unsigned int uiSum )
{
	unsigned short*	pusPtr;
	unsigned short	usThis, usChecksum;
	int				i;

	// calculate.

	pusPtr = (unsigned short*) ( pbStart + iOffset );
	for( i=0; i<iNumOfWords; i ++, pusPtr ++ )
	{
		usThis = ReadUShortFromPacket( (BYTE*) pusPtr );
		uiSum += usThis;
	}

	usChecksum = ChecksumFromSum( uiSum );

	// return.

	return usChecksum;
}

static unsigned int CalculateSumWithLenCheck( IN BYTE* pbStart, IN INT iOffset, IN INT iNumOfBytes )
{
	unsigned short*	pusPtr;
	int				i;
	unsigned int	uiSum;
	unsigned short	usThis;

	// calculate.

	uiSum = 0;
	pusPtr = (unsigned short*) ( pbStart + iOffset );
	for( i=0; i<iNumOfBytes/2; i ++, pusPtr ++ )
	{
		usThis = ReadUShortFromPacket( (BYTE*) pusPtr );
		uiSum += usThis;
	}

	if ( iNumOfBytes % 2 )
		uiSum += ( * ( pbStart + iOffset + iNumOfBytes - 1 ) ) << 8;

	return uiSum;
}

static VOID InitializeIpPseudoHeader( OUT BYTE* vbPseudoHeader, IN BYTE* pbNew, IN DWORD dwDataSize )
{
	// initialize.

	memcpy( vbPseudoHeader, pbNew + 26, 4 ); // source ip address.
	memcpy( vbPseudoHeader + 4, pbNew + 30, 4 ); // dest ip address.
	* ( vbPseudoHeader + 8 ) = 0; // zero.
	* ( vbPseudoHeader + 9 ) = * ( pbNew + 23 ); // protocol.
	WriteUShortToPacket( vbPseudoHeader + 10, (unsigned short) ( dwDataSize - 34 ) ); // TCP length.
}

VOID SendPacketHook( IN PNT_PROTOCOL pnpProtocol, IN PNT_OPEN_ADAPTER pnoaAdapter, IN BYTE* pbData, IN DWORD dwDataSize )
{
	PDEVICE_EXTENSION	extension = g_pDeviceObject->DeviceExtension;
	BYTE*				pbNew;
	BYTE*				pbMem;
	unsigned short		usEthertype;
	BYTE				vbNatMacAddress [] = { 0x02, 0x00, 0x4C, 0x4F, 0x4F, 0x32 };
	static BYTE			vbLoopbackMacAddress [] = { 0x02, 0x00, 0x4C, 0x4F, 0x4F, 0x50 }; // default one.
	static BYTE			vbBroadcastMacAddress [] = { 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF };
	BOOL				bSend = FALSE;
	int					iLen;
//	BYTE				vbPseudoHeader[ 3 * 4 ];
	KIRQL				irql;

	if ( extension->dwLtLoopbackAdapterIpAddress != 0 && extension->dwLtTrickNatIpAddress != 0 &&
		pnoaAdapter->pobBlockPtr && pnoaAdapter->pobBlockPtr->ReceiveHandler &&
		dwDataSize > 14
		&&
		( memcmp( pbData, & vbLoopbackMacAddress, 6 ) == 0 || // ethernet dest mac = loopback/NAT/Broadcast address.
		memcmp( pbData, & vbNatMacAddress , 6 ) == 0 ||
		memcmp( pbData, & vbBroadcastMacAddress , 6 ) == 0 )
		&&
		( memcmp( pbData + 6, & vbLoopbackMacAddress, 6 ) == 0 || // ethernet src mac = loopback/NAT address.
		memcmp( pbData + 6, & vbNatMacAddress , 6 ) == 0 )
		)
	{
		pbMem = (BYTE*) ExAllocatePool( NonPagedPool, dwDataSize + 12 /* TAG SIZE */ );
		if ( pbMem )
		{
			// copy the packet memory.

			memcpy( pbMem, "LOOPTRCK", 8 );
			* (DWORD*) ( pbMem + 8 ) = dwDataSize;
			memcpy( pbNew = pbMem + 12, pbData, dwDataSize );

			// manipulate the packet data.

			usEthertype = ReadUShortFromPacket( pbNew + 12 );
			switch ( usEthertype )
			{
			case 0x0806: // ARP
				{
					if ( dwDataSize >= 42 && // ARP "Request" Packet size.
						ReadUShortFromPacket( pbNew + 14 ) == 0x0001 && // Hardware type = Ethernet.
						ReadUShortFromPacket( pbNew + 16 ) == 0x0800 && // Protocol type = IP.
						* ( pbNew + 18 ) == 6 && // Hardware address length = 6.
						* ( pbNew + 19 ) == 4 && // Protocol address length = 4.
						ReadUShortFromPacket( pbNew + 20 ) == 0x0001 && // Opcode = Request.
						ReadUIntFromPacket( pbNew + 28 ) == extension->dwLtLoopbackAdapterIpAddress && // Source protocol address = loopback adapter assigned IP address.
						ReadUIntFromPacket( pbNew + 38 ) == extension->dwLtTrickNatIpAddress ) // Destination protocol address = NDIS Monitor NAT address.
					{
						// return an ARP response.
						memcpy( pbNew, pbData + 6, 6 ); // ethernet dest mac = request src mac.
						memcpy( pbNew + 6, vbNatMacAddress, 6 ); // ethernet source mac = NAT mac.
						WriteUShortToPacket( pbNew + 20, 0x0002 ); // Opcode = Response.
						memcpy( pbNew + 22, vbNatMacAddress, 6 ); // source hardware address = NAT mac.
						memcpy( pbNew + 32, pbData + 6, 6 ); // dest hardware address = request src mac.
						WriteUIntToPacket( pbNew + 28, extension->dwLtTrickNatIpAddress ); // source protocol address = NAT address.
						WriteUIntToPacket( pbNew + 38, extension->dwLtLoopbackAdapterIpAddress ); // dest protocol address = loopback address.

						// store the mac address of the loopback adapter.
						memcpy( & vbLoopbackMacAddress, pbData + 6, 6 );

						// can send.
						bSend = TRUE;
					}
				}
				break;

			case 0x0800: // IP
				{
					// check whether to accept the packet.

					if (
						dwDataSize >= 34 // packet size >= min ethernet II + IP header packet size.
					)
					{
						// reverse the mac addresses and ip addresses.

						if ( memcmp( pbNew, & vbBroadcastMacAddress , 6 ) == 0 )
						{
							if ( memcmp( pbNew + 6, & vbLoopbackMacAddress, 6 ) == 0 )
								memcpy( pbNew + 6, & vbNatMacAddress, 6 );
							else
								memcpy( pbNew + 6, & vbLoopbackMacAddress, 6 );
						}
						else
						{
							// reverse the mac addresses.

							memcpy( pbNew, pbData + 6, 6 );
							memcpy( pbNew + 6, pbData, 6 );

							// reverse the IP addresses.

							memcpy( pbNew + 26, pbData + 30, 4 );
							memcpy( pbNew + 30, pbData + 26, 4 );
						}

						// calculate the new header checksum.

						iLen = ( ( * ( pbNew + 14 ) ) & 0x0F ) * 2; // in 16bit words...
						if ( dwDataSize >= (DWORD) ( iLen * 2 + 14 ) )
						{
							// calculate and set checksum.

							WriteUShortToPacket( pbNew + 24, 0x0000 ); // reset chksum.
							WriteUShortToPacket( pbNew + 24, CalculateCheckSum( pbNew, 14, iLen, 0 ) ); // new checksum.

							// we can send the packet.

							bSend = TRUE;

							// take care of the transport protocol header.

							switch( * ( pbNew + 23 ) )
							{
							case 0x11: // UDP
								{
//									if ( dwDataSize >= 42 ) // min ethernet II + IP + UDP.
//									{
//										// calculate new checksum.
//
//										InitializeIpPseudoHeader( vbPseudoHeader, pbNew, dwDataSize );
//										WriteUShortToPacket( pbNew + 40, 0x0000 ); // reset chksum.
//										WriteUShortToPacket( pbNew + 40,
//											CalculateCheckSum( vbPseudoHeader, 0, sizeof( vbPseudoHeader ) / sizeof( unsigned short ),
//												CalculateSumWithLenCheck( pbNew, 34, dwDataSize - 34 ) ) );
//
//										if ( ReadUShortFromPacket( pbNew + 40 ) == 0x0000 ) // # specific of UDP #.
//											WriteUShortToPacket( pbNew + 40, 0xFFFF );
//									}
								}
								break;

							case 0x06: // TCP
								{
//									if ( dwDataSize >= 54 ) // min ethernet II + IP + TCP.
//									{
//										// calculate new checksum.
//
//										InitializeIpPseudoHeader( vbPseudoHeader, pbNew, dwDataSize );
//										WriteUShortToPacket( pbNew + 50, 0x0000 ); // reset chksum.
//										WriteUShortToPacket( pbNew + 50,
//											CalculateCheckSum( vbPseudoHeader, 0, sizeof( vbPseudoHeader ) / sizeof( unsigned short ),
//												CalculateSumWithLenCheck( pbNew, 34, dwDataSize - 34 ) ) );
//									}
								}
								break;

							case 0x01: // ICMP
								{
									if ( dwDataSize >= 38 &&
										* ( pbNew + 34 ) == 0x08 ) // Type = echo request.
									{
										// set the packet as a response.

										* ( pbNew + 34 ) = 0x00; // echo reply.

										// calculate new checksum.

										WriteUShortToPacket( pbNew + 36, 0x0000 );
										WriteUShortToPacket( pbNew + 36,
											ChecksumFromSum( CalculateSumWithLenCheck( pbNew, 34, dwDataSize - 34 ) ) );
									}
								}
								break;
							}
						}
					}
				}
				break;
			}

			// send the packet back to the protocol.

			if ( bSend )
			{
				KeRaiseIrql( DISPATCH_LEVEL, & irql );

					pnoaAdapter->pobBlockPtr->ReceiveHandler(
						pnoaAdapter->pobBlockPtr->ProtocolBindingContext,
						(NDIS_HANDLE) pbMem,
						pbNew, 14,
						pbNew + 14, dwDataSize - 14,
						dwDataSize - 14 );

				KeLowerIrql( irql );
			}

			// free memory.

			ExFreePool( pbMem );
		}
	}
}

//=====================================
// RetrievePacket Function Definition.
//=====================================

BOOL RetrievePacket( OUT NEXT_PACKET* pOut, OUT HOOK_STATS* pStats )
{
	PDEVICE_EXTENSION		extension = g_pDeviceObject->DeviceExtension;

	// get.

	BOOL		retval = FALSE;

	NdisAcquireSpinLock( & extension->PacketsBuffSpinLock );

		if ( extension->PacketsBuff && extension->PacketsBuffMaxItems && extension->PacketsBuffStart != -1 )
		{
			NEXT_PACKET*		pThis = & extension->PacketsBuff[ extension->PacketsBuffStart ];

			//
			// manage the circular buffer.
			//

			// inc start pos.

			extension->PacketsBuffStart ++;
			if ( extension->PacketsBuffStart == extension->PacketsBuffMaxItems )
				extension->PacketsBuffStart = 0;

			// check whether the list was emptied.

			if ( extension->PacketsBuffStart == extension->PacketsBuffEnd )
			{
				extension->PacketsBuffStart = -1;
				extension->PacketsBuffEnd = 0;
			}

			//
			// return this one.
			//

			* pOut = * pThis;
			pThis->pbData = NULL;

			//
			// complete.
			//

			retval = TRUE;

			pStats->dwLostPackets = extension->PacketsLost;
			pStats->bArePacketsRemaining = ( extension->PacketsBuffStart == -1 ) ? 0 : 1;

			extension->PacketsLost = 0;
		}

	NdisReleaseSpinLock( & extension->PacketsBuffSpinLock );

	// return.

	return retval;
}

//===========================
// packet utility functions.
//===========================

static BYTE* NdisPacket2MemoryRegion( IN PNDIS_PACKET Packet, IN BYTE* pbHeaderBuffer, IN ULONG ulHeaderBufferSize, OUT ULONG* pulSize )
{
	BYTE*			pbMem = NULL;
	UINT			uiTotalSize = 0;
	ULONG			ulMemSize = 0;
	PNDIS_BUFFER	pnbCurrent = NULL;
	PVOID			pvPtr = NULL;
	UINT			uiPtrDim = 0;
	UINT			uiCount = 0;

	* pulSize = 0;

	// # allocate the memory. #

	NdisQueryPacket( Packet,
		NULL, NULL,
		& pnbCurrent,
		& uiTotalSize );

	ulMemSize = ulHeaderBufferSize + uiTotalSize;
	if ( ulMemSize == 0 )
		return NULL;

	pbMem = (BYTE*) ExAllocatePool( NonPagedPool, ulMemSize );
	if ( pbMem == NULL )
		return NULL;

	memset( pbMem, 0xCC, ulMemSize ); // debug feature...

	// # fill the memory with packet data. #

	while( pnbCurrent )
	{
		NdisQueryBuffer( pnbCurrent, & pvPtr, & uiPtrDim );

		if ( pvPtr == NULL || uiCount + uiPtrDim > uiTotalSize )
		{
			ExFreePool( pbMem );
			return NULL;
		}
		else
		{
			memcpy( pbMem + uiCount, pvPtr, uiPtrDim );
			uiCount += uiPtrDim;
		}

		NdisGetNextBuffer( pnbCurrent, & pnbCurrent );
	}

	if ( uiCount != uiTotalSize )
	{
		ExFreePool( pbMem );
		return NULL;
	}

	// # check whether inserting the header is required. #

	* pulSize = uiTotalSize;

	if ( pbHeaderBuffer && ulHeaderBufferSize )
	{
		if ( memcmp( pbMem, pbHeaderBuffer, ulHeaderBufferSize ) )
		{
			// move the memory and copy the header.

			memmove( pbMem + ulHeaderBufferSize, pbMem, uiTotalSize );
			memcpy( pbMem, pbHeaderBuffer, ulHeaderBufferSize );

			* pulSize += ulHeaderBufferSize;
		}
	}

	// # return. #

	return pbMem;
}

static VOID IncrementPacketsLostCounter ()
{
	PDEVICE_EXTENSION		extension = g_pDeviceObject->DeviceExtension;

	NdisAcquireSpinLock( & extension->PacketsBuffSpinLock ); // more correct than InterlockedIncrement in this case.

		extension->PacketsLost ++;

	NdisReleaseSpinLock( & extension->PacketsBuffSpinLock );
}

static DWORD QueueReceiveWorkItem( IN PNT_PROTOCOL pnpProtocol, IN BYTE* pbHeader, IN UINT uiHeaderSize, IN PVOID pvPacketRef, IN DWORD dwAsyncOpAdapterOrd, IN DWORD dwSerial )
{
	PDEVICE_EXTENSION		extension = g_pDeviceObject->DeviceExtension;
	BYTE*					pbMem = NULL;
	DWORD					dwRetVal = -1;

	NdisAcquireSpinLock( & extension->ReceiveWorkItemSpinLock );

		if ( pnpProtocol->pbWorkItemHeader )
			ExFreePool( pnpProtocol->pbWorkItemHeader );

		pbMem = ExAllocatePool( NonPagedPool, uiHeaderSize );
		if ( pbMem )
			memcpy( pbMem, pbHeader, uiHeaderSize );
		else
			uiHeaderSize = 0;

		pnpProtocol->pbWorkItemHeader = pbMem;
		pnpProtocol->uiWorkItemHeaderSize = uiHeaderSize;
		dwRetVal = pnpProtocol->dwWorkItemID = ++ ( extension->ReceiveWorkItemIdCounter );

		pnpProtocol->pvWorkItemPacketRef = pvPacketRef;
		pnpProtocol->dwWorkItemAsyncOpAdapterOrd = dwAsyncOpAdapterOrd;

		pnpProtocol->dwWorkItemPacketSerial = dwSerial;

	NdisReleaseSpinLock( & extension->ReceiveWorkItemSpinLock );

	return dwRetVal;
}

static VOID RemoveReceiveWorkItem( IN PNT_PROTOCOL pnpProtocol, IN DWORD dwID )
{
	PDEVICE_EXTENSION		extension = g_pDeviceObject->DeviceExtension;

	NdisAcquireSpinLock( & extension->ReceiveWorkItemSpinLock );

		if ( pnpProtocol->pbWorkItemHeader && pnpProtocol->dwWorkItemID == dwID )
		{
			// free...

			ExFreePool( pnpProtocol->pbWorkItemHeader );

			pnpProtocol->pbWorkItemHeader = NULL;
			pnpProtocol->uiWorkItemHeaderSize = 0;

			pnpProtocol->dwWorkItemPacketSerial = (DWORD) -1;
		}

	NdisReleaseSpinLock( & extension->ReceiveWorkItemSpinLock );
}

static BOOL GetWorkItemData( IN PNT_PROTOCOL pnpProtocol, OUT BYTE** ppbWorkItemHeader, OUT UINT* puiWorkItemHeaderSize, OUT DWORD* pdwWorkItemID, IN BOOLEAN bRemove, OUT DWORD* pdwSerial )
{
	PDEVICE_EXTENSION		extension = g_pDeviceObject->DeviceExtension;

	NdisAcquireSpinLock( & extension->ReceiveWorkItemSpinLock );

		* ppbWorkItemHeader = pnpProtocol->pbWorkItemHeader;
		* puiWorkItemHeaderSize = pnpProtocol->uiWorkItemHeaderSize;
		* pdwWorkItemID = pnpProtocol->dwWorkItemID;
		* pdwSerial = pnpProtocol->dwWorkItemPacketSerial;

		if ( bRemove )
		{
			pnpProtocol->pbWorkItemHeader = NULL;
			pnpProtocol->uiWorkItemHeaderSize = 0;
		}

	NdisReleaseSpinLock( & extension->ReceiveWorkItemSpinLock );

	return ( * ppbWorkItemHeader ) != NULL;
}

static VOID SetWorkItemPacketRefAndAsyncInfo( IN PNT_PROTOCOL pnpProtocol, IN PVOID pvPacketRef, IN DWORD dwAsyncOpAdapterOrd )
{
	PDEVICE_EXTENSION		extension = g_pDeviceObject->DeviceExtension;

	NdisAcquireSpinLock( & extension->ReceiveWorkItemSpinLock );

		pnpProtocol->pvWorkItemPacketRef = pvPacketRef;
		pnpProtocol->dwWorkItemAsyncOpAdapterOrd = dwAsyncOpAdapterOrd;

	NdisReleaseSpinLock( & extension->ReceiveWorkItemSpinLock );
}

//============================================
// Intercept_SendHandler Function Definition.
//============================================

BOOLEAN Intercept_SendHandler( IN PNT_PROTOCOL pnpProtocol, IN PNT_OPEN_ADAPTER pnoaAdapter, OUT NDIS_STATUS* pRetStatus, IN OUT NDIS_HANDLE* pMacBindingHandle, IN OUT PNDIS_PACKET* pPacket )
{
	PDEVICE_EXTENSION		extension = g_pDeviceObject->DeviceExtension;

	// process.

	if ( HaveToProcess( pnoaAdapter ) )
	{
		BYTE*		pbPacket = NULL;
		ULONG		ulPacketSize = 0;

		// allocate the packet memory.

		pbPacket = ( * pPacket ) ? NdisPacket2MemoryRegion( * pPacket, NULL, 0, & ulPacketSize ) : NULL;

		if ( pbPacket )
		{
			QueuePacket( pnoaAdapter, pbPacket, ulPacketSize, 1, GetNextPacketSerial () );
			SendPacketHook( pnpProtocol, pnoaAdapter, pbPacket, ulPacketSize );
		}
		else
		{
			IncrementPacketsLostCounter ();
		}
	}

	// return. ALWAYS give control to the original handler.

	return TRUE;
}

//=========================================
// New_ReceiveHandler Function Definition.
//=========================================

#pragma optimize( "", off )

NDIS_STATUS __cdecl New_ReceiveHandler( IN OUT DWORD HookPrivateStorage, IN OUT DWORD CallingFnRetAddress, IN OUT NDIS_HANDLE ProtocolBindingContext, IN OUT NDIS_HANDLE MacReceiveContext, IN OUT PVOID HeaderBuffer, IN OUT UINT HeaderBufferSize, IN OUT PVOID LookAheadBuffer, IN OUT UINT LookaheadBufferSize, IN OUT UINT PacketSize )
{
	NDIS_STATUS				nsRetStatus = NDIS_STATUS_SUCCESS;
	RECEIVE_HANDLER			pfnOriginal = NULL;
	DWORD					OpenAdapterId = HookPrivateStorage;
	PDEVICE_EXTENSION		pdeExtension = g_pDeviceObject->DeviceExtension;
	BOOLEAN					bExecuteOriginalHandler = TRUE;

	// Get a pointer to the Original Handler.

	pfnOriginal = g_pntoaAdapters[ OpenAdapterId ].Original_ReceiveHandler;

	// Check if we have to intercept the NDIS Request.

	if ( pdeExtension->InterceptNDIS )
		bExecuteOriginalHandler = Intercept_ReceiveHandler(
			& g_pntpProtocols[ g_pntoaAdapters[ OpenAdapterId ].dwProtocolOrd ], & g_pntoaAdapters[ OpenAdapterId ],
			& nsRetStatus, & ProtocolBindingContext, & MacReceiveContext, & HeaderBuffer, & HeaderBufferSize, & LookAheadBuffer, & LookaheadBufferSize, & PacketSize,
			pfnOriginal );

	// Make call the Original Handler and return.

	if ( bExecuteOriginalHandler )
		HookPrivateStorage = (DWORD) pfnOriginal; // Original Handler.
	else
		HookPrivateStorage = 7 * 0x4; // Passed Parameters Size.

	return nsRetStatus;
}

#pragma optimize( "", on )

//===============================================
// Intercept_ReceiveHandler Function Definition.
//===============================================

BOOLEAN Intercept_ReceiveHandler( IN PNT_PROTOCOL pnpProtocol, IN PNT_OPEN_ADAPTER pnoaAdapter, OUT NDIS_STATUS* pRetStatus, IN OUT NDIS_HANDLE* pProtocolBindingContext, IN OUT NDIS_HANDLE* pMacReceiveContext, IN OUT PVOID* pHeaderBuffer, IN OUT UINT* pHeaderBufferSize, IN OUT PVOID* pLookAheadBuffer, IN OUT UINT* pLookaheadBufferSize, IN OUT UINT* pPacketSize, IN RECEIVE_HANDLER pfnOriginal )
{
	PDEVICE_EXTENSION		extension = g_pDeviceObject->DeviceExtension;
	NDIS_STATUS				ndsStatus;
	BOOLEAN					bNTD = FALSE;
	DWORD					dwID = -1;
	DWORD					dwSerial;

	if ( HaveToProcess( pnoaAdapter ) == FALSE )
		return TRUE; // let the original function be called.

	// ## generate the serial number for the packet. ##

	dwSerial = GetNextPacketSerial ();

	// ## call the original function. ##

	if ( * pPacketSize > * pLookaheadBufferSize ) // the protocol probably will call NdisTransferData...
	{
		// queue a work item about this operation.
		dwID = QueueReceiveWorkItem( pnpProtocol, * pHeaderBuffer, * pHeaderBufferSize, NULL, -1, dwSerial );

		// set the flag.
		bNTD = TRUE;
	}

	ndsStatus = pfnOriginal(
		* pProtocolBindingContext, * pMacReceiveContext, * pHeaderBuffer, * pHeaderBufferSize,
		* pLookAheadBuffer, * pLookaheadBufferSize, * pPacketSize );

	if ( ndsStatus == NDIS_STATUS_SUCCESS ) // the protocol accepted the packet...
	{
		if ( ! bNTD ) // copy the packet data in our buffers...
		{
			ULONG			ulSize = * pHeaderBufferSize + * pLookaheadBufferSize;
			if ( ulSize )
			{
				BYTE*		pbMem = (BYTE*) ExAllocatePool( NonPagedPool, ulSize );
				if ( pbMem == NULL )
				{
					IncrementPacketsLostCounter ();
				}
				else
				{
					memcpy( pbMem, * pHeaderBuffer, * pHeaderBufferSize );
					memcpy( pbMem + * pHeaderBufferSize, * pLookAheadBuffer, * pLookaheadBufferSize );

					QueuePacket( pnoaAdapter, pbMem, ulSize, 0, dwSerial );
				}
			}
		}
	}
	else // the protocol refused the packet...
	{
		if ( bNTD ) // remove the work item...
			RemoveReceiveWorkItem( pnpProtocol, dwID );
	}

	// ## prevent the original function from being called. ##

	* pRetStatus = ndsStatus;

	return FALSE;
}

//=================================================
// New_PostNt31ReceiveHandler Function Definition.
//=================================================

#pragma optimize( "", off )

NDIS_STATUS __cdecl New_PostNt31ReceiveHandler( IN OUT DWORD HookPrivateStorage, IN OUT DWORD CallingFnRetAddress, IN OUT NDIS_HANDLE ProtocolBindingContext, IN OUT NDIS_HANDLE MacReceiveContext, IN OUT PVOID HeaderBuffer, IN OUT UINT HeaderBufferSize, IN OUT PVOID LookAheadBuffer, IN OUT UINT LookaheadBufferSize, IN OUT UINT PacketSize )
{
	NDIS_STATUS				nsRetStatus = NDIS_STATUS_SUCCESS;
	RECEIVE_HANDLER			pfnOriginal = NULL;
	DWORD					OpenAdapterId = HookPrivateStorage;
	PDEVICE_EXTENSION		pdeExtension = g_pDeviceObject->DeviceExtension;
	BOOLEAN					bExecuteOriginalHandler = TRUE;

	// Get a pointer to the Original Handler.

	pfnOriginal = g_pntoaAdapters[ OpenAdapterId ].Original_PostNt31ReceiveHandler;

	// Check if we have to intercept the NDIS Request.

	if ( pdeExtension->InterceptNDIS )
		bExecuteOriginalHandler = Intercept_PostNt31ReceiveHandler(
			& g_pntpProtocols[ g_pntoaAdapters[ OpenAdapterId ].dwProtocolOrd ], & g_pntoaAdapters[ OpenAdapterId ],
			& nsRetStatus, & ProtocolBindingContext, & MacReceiveContext, & HeaderBuffer, & HeaderBufferSize, & LookAheadBuffer, & LookaheadBufferSize, & PacketSize,
			pfnOriginal );

	// Make call the Original Handler and return.

	if ( bExecuteOriginalHandler )
		HookPrivateStorage = (DWORD) pfnOriginal; // Original Handler.
	else
		HookPrivateStorage = 7 * 0x4; // Passed Parameters Size.

	return nsRetStatus;
}

#pragma optimize( "", on )

//=======================================================
// Intercept_PostNt31ReceiveHandler Function Definition.
//=======================================================

BOOLEAN Intercept_PostNt31ReceiveHandler( IN PNT_PROTOCOL pnpProtocol, IN PNT_OPEN_ADAPTER pnoaAdapter, OUT NDIS_STATUS* pRetStatus, IN OUT NDIS_HANDLE* pProtocolBindingContext, IN OUT NDIS_HANDLE* pMacReceiveContext, IN OUT PVOID* pHeaderBuffer, IN OUT UINT* pHeaderBufferSize, IN OUT PVOID* pLookAheadBuffer, IN OUT UINT* pLookaheadBufferSize, IN OUT UINT* pPacketSize, IN RECEIVE_HANDLER pfnOriginal )
{
	// forward.
	return Intercept_ReceiveHandler( pnpProtocol, pnoaAdapter, pRetStatus, pProtocolBindingContext, pMacReceiveContext, pHeaderBuffer, pHeaderBufferSize, pLookAheadBuffer, pLookaheadBufferSize, pPacketSize, pfnOriginal );
}

//=============================================
// New_SendPacketsHandler Function Definition.
//=============================================

#pragma optimize( "", off )

VOID __cdecl New_SendPacketsHandler( IN OUT DWORD HookPrivateStorage, IN OUT DWORD CallingFnRetAddress, IN OUT NDIS_HANDLE MiniportAdapterContext, IN OUT PPNDIS_PACKET PacketArray, IN OUT UINT NumberOfPackets )
{
	SEND_PACKETS_HANDLER	pfnOriginal = NULL;
	DWORD					OpenAdapterId = HookPrivateStorage;
	PDEVICE_EXTENSION		pdeExtension = g_pDeviceObject->DeviceExtension;
	BOOLEAN					bExecuteOriginalHandler = TRUE;

	// Get a pointer to the Original Handler.

	pfnOriginal = g_pntoaAdapters[ OpenAdapterId ].Original_SendPacketsHandler;

	// Check if we have to intercept the NDIS Request.

	if ( pdeExtension->InterceptNDIS )
		bExecuteOriginalHandler = Intercept_SendPacketsHandler(
			& g_pntpProtocols[ g_pntoaAdapters[ OpenAdapterId ].dwProtocolOrd ], & g_pntoaAdapters[ OpenAdapterId ],
			& MiniportAdapterContext, & PacketArray, & NumberOfPackets );

	// Make call the Original Handler and return.

	if ( bExecuteOriginalHandler )
		HookPrivateStorage = (DWORD) pfnOriginal; // Original Handler.
	else
		HookPrivateStorage = 3 * 0x4; // Passed Parameters Size.

	return;
}

#pragma optimize( "", on )

//===================================================
// Intercept_SendPacketsHandler Function Definition.
//===================================================

BOOLEAN Intercept_SendPacketsHandler( IN PNT_PROTOCOL pnpProtocol, IN PNT_OPEN_ADAPTER pnoaAdapter, IN OUT NDIS_HANDLE* pMiniportAdapterContext, IN OUT PPNDIS_PACKET* pPacketArray, IN OUT UINT* pNumberOfPackets )
{
	PDEVICE_EXTENSION		extension = g_pDeviceObject->DeviceExtension;
	int						i;
	BYTE*					pbPacket = NULL;
	ULONG					ulPacketSize = 0;

	// process.

	if ( HaveToProcess( pnoaAdapter ) &&
		( *pNumberOfPackets ) &&
		( *pPacketArray ) )
	{
		// Queue the various packets.

		for( i=0; i < (int) ( *pNumberOfPackets ); i ++ )
		{
			PNDIS_PACKET		pPacket = ( *pPacketArray )[ i ];

			if ( pPacket )
			{
				// allocate the packet memory.

				pbPacket = NdisPacket2MemoryRegion( pPacket, NULL, 0, & ulPacketSize );

				if ( pbPacket )
				{
					QueuePacket( pnoaAdapter, pbPacket, ulPacketSize, 1, GetNextPacketSerial () );
					SendPacketHook( pnpProtocol, pnoaAdapter, pbPacket, ulPacketSize );
				}
				else
				{
					IncrementPacketsLostCounter ();
				}
			}
		}
	}

	// Return.

	return TRUE; // let the original function be called.
}

//==============================================
// New_TransferDataHandler Function Definition.
//==============================================

#pragma optimize( "", off )

NDIS_STATUS __cdecl New_TransferDataHandler( IN OUT DWORD HookPrivateStorage, IN OUT DWORD CallingFnRetAddress, IN OUT NDIS_HANDLE MacBindingHandle, IN OUT NDIS_HANDLE MacReceiveContext, IN OUT UINT ByteOffset, IN OUT UINT BytesToTransfer, IN OUT PNDIS_PACKET Packet, IN OUT PUINT BytesTransferred )
{
	NDIS_STATUS				nsRetStatus = NDIS_STATUS_SUCCESS;
	TRANSFER_DATA_HANDLER	pfnOriginal = NULL;
	DWORD					OpenAdapterId = HookPrivateStorage;
	PDEVICE_EXTENSION		pdeExtension = g_pDeviceObject->DeviceExtension;
	BOOLEAN					bExecuteOriginalHandler = TRUE;

	// Get a pointer to the Original Handler.

	pfnOriginal = g_pntoaAdapters[ OpenAdapterId ].Original_TransferDataHandler;

	// Check if we have to intercept the NDIS Request.

	if ( pdeExtension->InterceptNDIS )
		bExecuteOriginalHandler = Intercept_TransferDataHandler(
			& g_pntpProtocols[ g_pntoaAdapters[ OpenAdapterId ].dwProtocolOrd ], & g_pntoaAdapters[ OpenAdapterId ],
			& nsRetStatus, & MacBindingHandle, & MacReceiveContext, & ByteOffset, & BytesToTransfer, & Packet, & BytesTransferred,
			pfnOriginal );

	// Make call the Original Handler and return.

	if ( bExecuteOriginalHandler )
		HookPrivateStorage = (DWORD) pfnOriginal; // Original Handler.
	else
		HookPrivateStorage = 6 * 0x4; // Passed Parameters Size.

	return nsRetStatus;
}

#pragma optimize( "", on )

//====================================================
// Intercept_TransferDataHandler Function Definition.
//====================================================

BOOLEAN Intercept_TransferDataHandler( IN PNT_PROTOCOL pnpProtocol, IN PNT_OPEN_ADAPTER pnoaAdapter, OUT NDIS_STATUS* pRetStatus, IN OUT NDIS_HANDLE* pMacBindingHandle, IN OUT NDIS_HANDLE* pMacReceiveContext, IN OUT UINT* pByteOffset, IN OUT UINT* pBytesToTransfer, IN OUT PNDIS_PACKET* pPacket, IN OUT PUINT* pBytesTransferred, IN TRANSFER_DATA_HANDLER pfnOriginal )
{
	PDEVICE_EXTENSION		extension = g_pDeviceObject->DeviceExtension;
	NDIS_STATUS				ndsStatus;
	BYTE*					pbHeader = NULL;
	UINT					uiHeaderSize = 0;
	DWORD					dwID = -1;
	BYTE*					pbAuxBuffer = NULL;
	UINT					uiAuxBufferSize = 0;
	BYTE*					pbPacket = NULL;
	ULONG					ulPacketSize = 0;
	BYTE*					pbBuffer;
	DWORD					dwBufferSize;
	PNDIS_BUFFER			pnbCurrent = NULL;
	UINT					uiTotalSize = 0;
	PVOID					pvPtr = NULL;
	UINT					uiPtrDim = 0;
	UINT					uiBytesTransferred, uiOffset, uiMemCpyDim;
	DWORD					dwSerial = (DWORD) -1;

	// Check to see whether the call originates in the context of a loopback trick send.

	if ( *pMacReceiveContext != NULL &&
		MmIsAddressValid( (BYTE*) *pMacReceiveContext ) && MmIsAddressValid( (BYTE*) *pMacReceiveContext + 8 ) &&
		memcmp( (VOID*) *pMacReceiveContext, "LOOPTRCK", 8 ) == 0 )
	{
		// ### get packet data. ###

		pbBuffer = (BYTE*) ( *pMacReceiveContext ) + 12 /* tag size */ + 14 /* ethernet II header size */;
		dwBufferSize = * (DWORD*)  ( (BYTE*) ( *pMacReceiveContext ) + 8 ) - 14 /* ethernet II header size */;

		// ### write in the packet. ###

		// get info on packet.

		NdisQueryPacket( *pPacket,
			NULL, NULL,
			& pnbCurrent,
			& uiTotalSize );

		// loop through packet buffers.

		uiOffset = *pByteOffset;
		uiBytesTransferred = 0;

		while( pnbCurrent )
		{
			// query buffer ptr.

			NdisQueryBuffer( pnbCurrent, & pvPtr, & uiPtrDim );
			if ( pvPtr == NULL )
				break;

			// write in this buffer.

			if ( uiOffset < dwBufferSize && uiBytesTransferred < *pBytesToTransfer )
			{
				// get block dim.

				uiMemCpyDim = uiPtrDim;

				if ( uiOffset + uiMemCpyDim > dwBufferSize )
					uiMemCpyDim = dwBufferSize - uiOffset;

				if ( uiBytesTransferred + uiMemCpyDim > *pBytesToTransfer )
					uiMemCpyDim = *pBytesToTransfer - uiBytesTransferred;

				// copy bytes.

				memcpy( pvPtr, pbBuffer + uiOffset, uiMemCpyDim );

				// update counter.

				uiBytesTransferred += uiMemCpyDim;
			}

			uiOffset += uiPtrDim;

			// get next packet buffer.

			NdisGetNextBuffer( pnbCurrent, & pnbCurrent );
		}

		// set num of bytes transferred.

		* ( *pBytesTransferred ) = uiBytesTransferred;

		// return success.

		* pRetStatus = NDIS_STATUS_SUCCESS;
		return FALSE;
	}

	// get wi data.

	if ( GetWorkItemData( pnpProtocol, & pbHeader, & uiHeaderSize, & dwID, FALSE /* DO NOT REMOVE IT */, & dwSerial ) == FALSE )
		return TRUE; // let the original function be called.

	// make a copy of the data.

	pbAuxBuffer = (BYTE*) ExAllocatePool( NonPagedPool, uiHeaderSize ); // we need a copy of the header.
	if ( pbAuxBuffer == NULL )
	{
		RemoveReceiveWorkItem( pnpProtocol, dwID );
		IncrementPacketsLostCounter ();
		return TRUE; // let the original function be called.
	}

	memcpy( pbAuxBuffer, pbHeader, uiHeaderSize );
	uiAuxBufferSize = uiHeaderSize;

	// set the packet ref pointer.

	SetWorkItemPacketRefAndAsyncInfo( pnpProtocol, (PVOID) * pPacket, pnoaAdapter->dwOrdinal );

	// call original fn.

	ndsStatus = pfnOriginal(
		* pMacBindingHandle, * pMacReceiveContext, * pByteOffset, * pBytesToTransfer, * pPacket, * pBytesTransferred );

	if ( ndsStatus == NDIS_STATUS_SUCCESS ) // ok, fetch the packet data and store it...
	{
		// fetch and store the packet.

		pbPacket = ( * pPacket ) ?
			NdisPacket2MemoryRegion( * pPacket, pbAuxBuffer, uiAuxBufferSize, & ulPacketSize )
			: NULL;

		if ( pbPacket )
			QueuePacket( pnoaAdapter, pbPacket, ulPacketSize, 0, dwSerial );
		else
			IncrementPacketsLostCounter ();

		// remove the work item.

		RemoveReceiveWorkItem( pnpProtocol, dwID );
	}
	else if ( ndsStatus == NDIS_STATUS_PENDING ) // wait until the completion routine is called...
	{
		//
		// ### do nothing, leave the work item at its original place ###
		//
	}
	else // error, remove the work item...
	{
		// remove the work item.

		RemoveReceiveWorkItem( pnpProtocol, dwID );
	}

	// free aux mem.

	ExFreePool( pbAuxBuffer );

	// ## prevent the original function from being called. ##

	* pRetStatus = ndsStatus;

	return FALSE;
}

//===============================================
// New_ReceivePacketHandler Function Definition.
//===============================================

#pragma optimize( "", off )

INT __cdecl New_ReceivePacketHandler( IN OUT DWORD HookPrivateStorage, IN OUT DWORD CallingFnRetAddress, IN OUT NDIS_HANDLE ProtocolBindingContext, IN OUT PNDIS_PACKET Packet )
{
	INT						iRetStatus = 0;
	RECEIVE_PACKET_HANDLER	pfnOriginal = NULL;
	DWORD					OpenAdapterId = HookPrivateStorage;
	PDEVICE_EXTENSION		pdeExtension = g_pDeviceObject->DeviceExtension;
	BOOLEAN					bExecuteOriginalHandler = TRUE;

	// Get a pointer to the Original Handler.

	pfnOriginal = g_pntoaAdapters[ OpenAdapterId ].Original_ReceivePacketHandler;

	// Check if we have to intercept the NDIS Request.

	if ( pdeExtension->InterceptNDIS )
		bExecuteOriginalHandler = Intercept_ReceivePacketHandler(
			& g_pntpProtocols[ g_pntoaAdapters[ OpenAdapterId ].dwProtocolOrd ], & g_pntoaAdapters[ OpenAdapterId ],
			& iRetStatus,
			& ProtocolBindingContext, & Packet,
			pfnOriginal );

	// Make call the Original Handler and return.

	if ( bExecuteOriginalHandler )
		HookPrivateStorage = (DWORD) pfnOriginal; // Original Handler.
	else
		HookPrivateStorage = 2 * 0x4; // Passed Parameters Size.

	return iRetStatus;
}

#pragma optimize( "", on )

//=====================================================
// Intercept_ReceivePacketHandler Function Definition.
//=====================================================

BOOLEAN Intercept_ReceivePacketHandler( IN PNT_PROTOCOL pnpProtocol, IN PNT_OPEN_ADAPTER pnoaAdapter, OUT INT* pRetStatus, IN OUT NDIS_HANDLE* pProtocolBindingContext, IN OUT PNDIS_PACKET* pPacket, IN RECEIVE_PACKET_HANDLER pfnOriginal )
{
	PDEVICE_EXTENSION		extension = g_pDeviceObject->DeviceExtension;
	INT						iRetVal = 0;
	BYTE*					pbPacket = NULL;
	ULONG					ulPacketSize = 0;
	DWORD					dwSerial;

	if ( HaveToProcess( pnoaAdapter ) == FALSE )
		return TRUE; // let the original function be called.

	// generate serial.

	dwSerial = GetNextPacketSerial ();

	// grab the packet data before the function can call NdisReturnPackets.

	pbPacket = ( * pPacket ) ?
		NdisPacket2MemoryRegion( * pPacket, NULL, 0, & ulPacketSize )
		: NULL;

	if ( pbPacket == NULL )
	{
		IncrementPacketsLostCounter ();
		return TRUE; // let the original function be called.
	}

	// call the original function.

	iRetVal = pfnOriginal( * pProtocolBindingContext, * pPacket );

	if ( /*iRetVal*/ TRUE /*=VP=08MAR2005*/ ) // # the protocol retained the packet.
		QueuePacket( pnoaAdapter, pbPacket, ulPacketSize, 0, dwSerial );
	else // # perhaps the protocol refused the packet.
		ExFreePool( pbPacket );

	// return.

	* pRetStatus = iRetVal;
	return FALSE; // don't call the original function.
}

//============================================
// New_TDCompleteHandler Function Definition.
//============================================

#pragma optimize( "", off )

VOID __cdecl New_TDCompleteHandler( IN OUT DWORD HookPrivateStorage, IN OUT DWORD CallingFnRetAddress, IN OUT NDIS_HANDLE MiniportAdapterHandle, IN OUT PNDIS_PACKET Packet, IN OUT NDIS_STATUS Status, IN OUT UINT BytesTransferred )
{
	NDIS_M_TD_COMPLETE_HANDLER	pfnOriginal = NULL;
	DWORD						OpenAdapterId = HookPrivateStorage;
	PDEVICE_EXTENSION			pdeExtension = g_pDeviceObject->DeviceExtension;
	BOOLEAN						bExecuteOriginalHandler = TRUE;

	// Get a pointer to the Original Handler.

	pfnOriginal = g_pntoaAdapters[ OpenAdapterId ].Original_TDCompleteHandler;

	// Check if we have to intercept the NDIS Request.

	if ( pdeExtension->InterceptNDIS )
		bExecuteOriginalHandler = Intercept_TDCompleteHandler(
			& g_pntpProtocols[ g_pntoaAdapters[ OpenAdapterId ].dwProtocolOrd ], & g_pntoaAdapters[ OpenAdapterId ],
			& MiniportAdapterHandle, & Packet, & Status, & BytesTransferred );			

	// Make call the Original Handler and return.

	if ( bExecuteOriginalHandler )
		HookPrivateStorage = (DWORD) pfnOriginal; // Original Handler.
	else
		HookPrivateStorage = 4 * 0x4; // Passed Parameters Size.

	return;
}

#pragma optimize( "", on )

//===================================================
// Intercept_TDCompleteHandler Function Definition.
//===================================================

BOOLEAN Intercept_TDCompleteHandler( IN PNT_PROTOCOL pnpProtocol, IN PNT_OPEN_ADAPTER pnoaAdapter, IN OUT NDIS_HANDLE* pMiniportAdapterHandle, IN OUT PNDIS_PACKET* pPacket, IN OUT NDIS_STATUS* pStatus, IN OUT UINT* pBytesTransferred )
{
	PDEVICE_EXTENSION	extension = g_pDeviceObject->DeviceExtension;
	int					i, c, j, k;
	NT_PROTOCOL*		pBase = NULL;
	NT_PROTOCOL*		pThis = NULL;
	NT_OPEN_ADAPTER*	pAdapters = NULL;
	BOOLEAN				found;
	BYTE*				pbHeader = NULL;
	UINT				uiHeaderSize = 0;
	DWORD				dwSerial = (DWORD) -1;

	// Discover the true protocol that initiated this async request.

	if ( extension->NdisProtocolList == NULL ||
		extension->NdisProtocolList->dwProtocolsNum == 0 )
	{
		return TRUE; // Return, letting the original function be called.
	}

	NdisAcquireSpinLock( & extension->ReceiveWorkItemSpinLock );

		pBase = (NT_PROTOCOL*) ( (BYTE*) extension->NdisProtocolList + sizeof( NT_PROTOCOL_LIST ) );
		c = extension->NdisProtocolList->dwProtocolsNum;

		for ( i=0; i<c; i ++ )
		{
			pThis = & pBase[ i ];

			//
			// compare the PACKET pointer...
			//

			if ( pThis->pbWorkItemHeader && pThis->pvWorkItemPacketRef == (PVOID) * pPacket )
			{
				// save the pointers.

				pbHeader = pThis->pbWorkItemHeader;
				uiHeaderSize = pThis->uiWorkItemHeaderSize;

				// save the packet serial.

				dwSerial = pThis->dwWorkItemPacketSerial;

				// release the work item.

				pThis->pbWorkItemHeader = NULL;
				pThis->uiWorkItemHeaderSize = 0;

				// set the correct "pnoaAdapter" pointer.

				found = FALSE;

				pAdapters = (NT_OPEN_ADAPTER*) ( (BYTE*) pBase + c * sizeof( NT_PROTOCOL ) );
				k = extension->NdisProtocolList->dwOpenAdaptersNum;

				for ( j=0; j<k; j ++ )
				{
					NT_OPEN_ADAPTER*		pA = & pAdapters[ j ];

					if ( pA->dwProtocolOrd == pThis->dwOrdinal &&
						pA->dwOrdinal == pThis->dwWorkItemAsyncOpAdapterOrd )
					{
						pnoaAdapter = pA;

						found = TRUE;
						break;
					}
				}

				if ( found == FALSE ) // ### THIS SHOULD NEVER HAPPEN ! ###
				{
					ExFreePool( pbHeader );

					pbHeader = NULL;
					uiHeaderSize = 0;

					IncrementPacketsLostCounter ();
				}

				// exit.

				break;
			}
		}

	NdisReleaseSpinLock( & extension->ReceiveWorkItemSpinLock );

	// Store the packet.

	if ( pbHeader && uiHeaderSize )
	{
		// store.

		if ( * pStatus == NDIS_STATUS_SUCCESS )
		{
			BYTE*			pbPacket = NULL;
			ULONG			ulPacketSize = 0;

			pbPacket = ( * pPacket ) ?
				NdisPacket2MemoryRegion( * pPacket, pbHeader, uiHeaderSize, & ulPacketSize )
				: NULL;

			if ( pbPacket )
				QueuePacket( pnoaAdapter, pbPacket, ulPacketSize, 0, dwSerial );
			else
				IncrementPacketsLostCounter ();
		}

		// free.

		ExFreePool( pbHeader );
	}

	// Return, letting the original function be called.

	return TRUE;
}

//=============================================
// HookInstalledProtocols Function Definition.
//=============================================

static BYTE*		g_pbCanHookRefMemStart = NULL;
static BYTE*		g_pbCanHookRefMemEnd = NULL;

static BOOLEAN CanHook( IN PVOID fnptr )
{
	if ( fnptr == NULL || g_pbCanHookRefMemStart == NULL || g_pbCanHookRefMemEnd == NULL ) // # pointer problem...
		return FALSE;
	else if ( ((BYTE*)fnptr) >= g_pbCanHookRefMemStart && ((BYTE*)fnptr) <= g_pbCanHookRefMemEnd ) // # already hooked...
		return FALSE;
	else // # ok, can hook...
		return TRUE;
}

static void FillStubCode( PNDISHOOK_HANDLER_STUB pnhhsThisStub, DWORD dwThisAdapterOrd, DWORD dwAddInfo )
{
	// Fill the Stub structure.

	pnhhsThisStub->bPushImm32Opcode = 0x68;
	pnhhsThisStub->dwOperationID = ( dwThisAdapterOrd << 0x10 ) | dwAddInfo;
	pnhhsThisStub->bJmpRel32Opcode = 0xE9;
	pnhhsThisStub->dwJmpDisplacement =
		(DWORD) & HandlerGeneralDispatcher -
		( (DWORD) pnhhsThisStub + FIELD_OFFSET( NDISHOOK_HANDLER_STUB, bJmpRel32Opcode ) ) -
		5;
}

NTSTATUS HookInstalledProtocols( OUT PNT_PROTOCOL_LIST* ppplProtocolList, OUT PNDISHOOK_HANDLER_STUB* ppnhhsStubs, OUT DWORD* pdwStubsNum, IN NDIS_HANDLE hFakeProtocolHandle )
{
	NDIS_HANDLE							hProtocol;
	NDIS_OPEN_BLOCK*					pobAdapter;
	NT_PROTOCOL_LIST					ntplHeader;
	PNT_PROTOCOL						pntpProtocols = NULL;
	PNT_OPEN_ADAPTER					pntoaAdapters = NULL;
	DWORD								dwProtocolsLimit, dwAdaptersLimit;
	DWORD								dwProtocolListSize;
	PNT_PROTOCOL_LIST					pplProtocolListTemp = NULL;
	NTSTATUS							nsReturnValue = STATUS_UNSUCCESSFUL;
	BYTE*								pbBytePtr;
	BOOLEAN								bFinalize = TRUE;
	PNT_PROTOCOL						pntpThisProtocol;
	DWORD								dwThisProtocolOrd;
	NDIS50_PROTOCOL_CHARACTERISTICS*	pn50pcProtChrs;
	PNT_OPEN_ADAPTER					pntoaThisAdapter;
	DWORD								dwThisAdapterOrd;
	PNDISHOOK_HANDLER_STUB				pnhhsStubsTemp = NULL;
	DWORD								dwStubsLimit, dwStubsPos = 0;
	PNDISHOOK_HANDLER_STUB				pnhhsThisStub;
	NDIS_MINIPORT_BLOCK*				pmbMiniport;

	* ppplProtocolList = NULL;
	* ppnhhsStubs = NULL;
	* pdwStubsNum = 0;

	__try
	{
		// Initialize the required memory.

		memset( & ntplHeader, 0, sizeof( ntplHeader ) );

		pntpProtocols = (PNT_PROTOCOL) ExAllocatePool( PagedPool, TEMPORARY_MEMORYBUFFER_SIZE );
		if ( pntpProtocols == NULL )
			return STATUS_UNSUCCESSFUL;

		memset( pntpProtocols, 0, TEMPORARY_MEMORYBUFFER_SIZE );

		pntoaAdapters = (PNT_OPEN_ADAPTER) ExAllocatePool( PagedPool, TEMPORARY_MEMORYBUFFER_SIZE );
		if ( pntoaAdapters == NULL )
			return STATUS_UNSUCCESSFUL;

		memset( pntoaAdapters, 0, TEMPORARY_MEMORYBUFFER_SIZE );

		pnhhsStubsTemp = (PNDISHOOK_HANDLER_STUB) ExAllocatePool( NonPagedPoolCacheAligned, STUBSLIST_BUFFER_MAXSIZE );
		if ( pnhhsStubsTemp == NULL )
			return STATUS_UNSUCCESSFUL;

		memset( pnhhsStubsTemp, 0, STUBSLIST_BUFFER_MAXSIZE );

		// Store the pointer for the "CanHook" function...

		g_pbCanHookRefMemStart = (BYTE*) pnhhsStubsTemp;
		g_pbCanHookRefMemEnd = g_pbCanHookRefMemStart + STUBSLIST_BUFFER_MAXSIZE;

		// Calculate the maximum number of elements that fit in the allocated memory.

		dwProtocolsLimit = ( TEMPORARY_MEMORYBUFFER_SIZE / sizeof( NT_PROTOCOL ) ) - 1;
		dwAdaptersLimit = ( TEMPORARY_MEMORYBUFFER_SIZE / sizeof( NT_OPEN_ADAPTER ) ) - 1;
		dwStubsLimit = ( STUBSLIST_BUFFER_MAXSIZE / sizeof( NDISHOOK_HANDLER_STUB ) ) - 1;

		// Store several global pointers in order to make the tables available to the Handler Dispatcher.

		g_pntpProtocols = pntpProtocols;
		g_pntoaAdapters = pntoaAdapters;

		// Iterate through the list of Protocols.

		__try
		{
			// Loop in the list of Installed Protocols.

			hProtocol = hFakeProtocolHandle;
			while( hProtocol = GetNextProtocol( hProtocol ) )
			{
				pn50pcProtChrs = GetProtocolCharacteristics( hProtocol );

				// Fill the structure data for this Protocol.

				if ( ntplHeader.dwProtocolsNum + 1 >= dwProtocolsLimit )
					return STATUS_UNSUCCESSFUL;

				dwThisProtocolOrd = ntplHeader.dwProtocolsNum ++;
				pntpThisProtocol = & pntpProtocols[ dwThisProtocolOrd ];

				memset( pntpThisProtocol, 0, sizeof( NT_PROTOCOL ) ); // # initialize to zeroes !! #

				pntpThisProtocol->dwOrdinal = dwThisProtocolOrd;
				pntpThisProtocol->hHandle = hProtocol;

				if ( pn50pcProtChrs->Name.Length + sizeof( WCHAR ) <= sizeof( pntpThisProtocol->szName ) )
				{
					memcpy( pntpThisProtocol->szName, pn50pcProtChrs->Name.Buffer, pn50pcProtChrs->Name.Length );
					pntpThisProtocol->szName[ pn50pcProtChrs->Name.Length / sizeof( WCHAR ) ] = L'\0';
				}
				else
				{
					wcscpy( pntpThisProtocol->szName, L"... Name Too Long ..." );
				}

				// Loop in the list of Protocol Open Adapters.

				pobAdapter = GetOpenQueue( hProtocol );
				if ( pobAdapter )
				{
					do
					{
						BOOLEAN			bHook = TRUE;
						WCHAR			szMedia[ 128 ] = L" { media = Unknown }";

						// Check whether WAN or not.

						pmbMiniport = GetMiniportBlock( pobAdapter );
						if ( pmbMiniport == NULL ||
							( pmbMiniport->MediaType != NdisMedium802_3 && pmbMiniport->MediaType != NdisMediumDix ) )
						{
							bHook = FALSE;
						}

						// Set media string.

						if ( pmbMiniport )
							switch( pmbMiniport->MediaType )
							{
							case NdisMedium802_3:
								wcscpy( szMedia, L" { media = Ethernet (802.3) }" );
								break;
							case NdisMedium802_5:
								wcscpy( szMedia, L" { media = Token Ring (802.5) }" );
								break;
							case NdisMediumFddi:
								wcscpy( szMedia, L" { media = Fiber Distributed Data Interface (FDDI) }" );
								break;
							case NdisMediumWan:
								wcscpy( szMedia, L" { media = Wide Area Network }" );
								break;
							case NdisMediumLocalTalk:
								wcscpy( szMedia, L" { media = LocalTalk }" );
								break;
							case NdisMediumDix:
								wcscpy( szMedia, L" { media = DIX Ethernet }" );
								break;
							case NdisMediumArcnetRaw:
								wcscpy( szMedia, L" { media = ARCNET }" );
								break;
							case NdisMediumArcnet878_2:
								wcscpy( szMedia, L" { media = ARCNET (878.2) }" );
								break;
							case NdisMediumAtm:
								wcscpy( szMedia, L" { media = ATM }" );
								break;
							case NdisMediumWirelessWan:
								wcscpy( szMedia, L" { media = Wireless }" );
								break;
							case NdisMediumIrda:
								wcscpy( szMedia, L" { media = Infrared (IrDA) }" );
								break;
							case NdisMediumCoWan:
								wcscpy( szMedia, L" { media = Connection-Oriented WAN }" );
								break;
							case NdisMedium1394:
								wcscpy( szMedia, L" { media = IEEE 1394 (Fire Wire) }" );
								break;
							}

						// Fill the structure data for this Adapter.

						if ( ntplHeader.dwOpenAdaptersNum + 1 >= dwAdaptersLimit )
							return STATUS_UNSUCCESSFUL;

						dwThisAdapterOrd = ntplHeader.dwOpenAdaptersNum ++;
						pntoaThisAdapter = & pntoaAdapters[ dwThisAdapterOrd ];

						pntoaThisAdapter->dwOrdinal = dwThisAdapterOrd;
						pntoaThisAdapter->dwProtocolOrd = dwThisProtocolOrd;

						pntoaThisAdapter->pobBlockPtr = pobAdapter;
						pntoaThisAdapter->pmbMiniportPtr = NULL;

						if ( pobAdapter->BindDeviceName != NULL )
						{
							if ( pobAdapter->BindDeviceName->Length + sizeof( WCHAR ) + wcslen( szMedia ) <= sizeof( pntoaThisAdapter->szName ) )
							{
								memcpy( pntoaThisAdapter->szName, pobAdapter->BindDeviceName->Buffer, pobAdapter->BindDeviceName->Length );
								pntoaThisAdapter->szName[ pobAdapter->BindDeviceName->Length / sizeof( WCHAR ) ] = L'\0';
							}
							else
							{
								wcscpy( pntoaThisAdapter->szName, L"... Name Too Long ..." );
							}
						}
						else
						{
							wcscpy( pntoaThisAdapter->szName, L"... Name Not Specified ..." );
						}

						wcscat( pntoaThisAdapter->szName, szMedia );

						//
						// ### Hook the Adapter Handlers. ###
						//

						// mem checkpoint.

						if ( dwStubsPos + 16 >= dwStubsLimit )
							return STATUS_UNSUCCESSFUL;

						// hook send handler.

						if ( bHook && CanHook( pobAdapter->SendHandler ) )
						{
							pnhhsThisStub = & pnhhsStubsTemp[ dwStubsPos ++ ];
							FillStubCode( pnhhsThisStub, dwThisAdapterOrd, SENDHANDLER_FNID );
							pntoaThisAdapter->Original_SendHandler = pobAdapter->SendHandler;
							pntoaThisAdapter->Stub_SendHandler = pnhhsThisStub;
							pobAdapter->SendHandler = (SEND_HANDLER) pnhhsThisStub;
						}
						else
						{
							pntoaThisAdapter->Original_SendHandler = NULL;
							pntoaThisAdapter->Stub_SendHandler = NULL;
						}

						// hook receive handler.

						if ( bHook && CanHook( pobAdapter->ReceiveHandler ) )
						{
							pnhhsThisStub = & pnhhsStubsTemp[ dwStubsPos ++ ];
							FillStubCode( pnhhsThisStub, dwThisAdapterOrd, RECEIVEHANDLER_FNID );
							pntoaThisAdapter->Original_ReceiveHandler = pobAdapter->ReceiveHandler;
							pntoaThisAdapter->Stub_ReceiveHandler = pnhhsThisStub;
							pobAdapter->ReceiveHandler = (RECEIVE_HANDLER) pnhhsThisStub;
						}
						else
						{
							pntoaThisAdapter->Original_ReceiveHandler = NULL;
							pntoaThisAdapter->Stub_ReceiveHandler = NULL;
						}

						// hook postnt31 receive handler.

						if ( bHook && CanHook( pobAdapter->PostNt31ReceiveHandler ) )
						{
							pnhhsThisStub = & pnhhsStubsTemp[ dwStubsPos ++ ];
							FillStubCode( pnhhsThisStub, dwThisAdapterOrd, POSTNT31RECEIVEHANDLER_FNID );
							pntoaThisAdapter->Original_PostNt31ReceiveHandler = pobAdapter->PostNt31ReceiveHandler;
							pntoaThisAdapter->Stub_PostNt31ReceiveHandler = pnhhsThisStub;
							pobAdapter->PostNt31ReceiveHandler = (RECEIVE_HANDLER) pnhhsThisStub;
						}
						else
						{
							pntoaThisAdapter->Original_PostNt31ReceiveHandler = NULL;
							pntoaThisAdapter->Stub_PostNt31ReceiveHandler = NULL;
						}

						// hook send packets handler.

						if ( bHook && CanHook( pobAdapter->SendPacketsHandler ) )
						{
							pnhhsThisStub = & pnhhsStubsTemp[ dwStubsPos ++ ];
							FillStubCode( pnhhsThisStub, dwThisAdapterOrd, SENDPACKETSHANDLER_FNID );
							pntoaThisAdapter->Original_SendPacketsHandler = pobAdapter->SendPacketsHandler;
							pntoaThisAdapter->Stub_SendPacketsHandler = pnhhsThisStub;
							pobAdapter->SendPacketsHandler = (SEND_PACKETS_HANDLER) pnhhsThisStub;
						}
						else
						{
							pntoaThisAdapter->Original_SendPacketsHandler = NULL;
							pntoaThisAdapter->Stub_SendPacketsHandler = NULL;
						}

						// hook transfer data handler.

						if ( bHook && CanHook( pobAdapter->TransferDataHandler ) )
						{
							pnhhsThisStub = & pnhhsStubsTemp[ dwStubsPos ++ ];
							FillStubCode( pnhhsThisStub, dwThisAdapterOrd, TRANSFERDATAHANDLER_FNID );
							pntoaThisAdapter->Original_TransferDataHandler = pobAdapter->TransferDataHandler;
							pntoaThisAdapter->Stub_TransferDataHandler = pnhhsThisStub;
							pobAdapter->TransferDataHandler = (TRANSFER_DATA_HANDLER) pnhhsThisStub;
						}
						else
						{
							pntoaThisAdapter->Original_TransferDataHandler = NULL;
							pntoaThisAdapter->Stub_TransferDataHandler = NULL;
						}

						// hook receive packet handler.

						if ( bHook && CanHook( pobAdapter->ReceivePacketHandler ) )
						{
							pnhhsThisStub = & pnhhsStubsTemp[ dwStubsPos ++ ];
							FillStubCode( pnhhsThisStub, dwThisAdapterOrd, RECEIVEPACKETHANDLER_FNID );
							pntoaThisAdapter->Original_ReceivePacketHandler = pobAdapter->ReceivePacketHandler;
							pntoaThisAdapter->Stub_ReceivePacketHandler = pnhhsThisStub;
							pobAdapter->ReceivePacketHandler = (RECEIVE_PACKET_HANDLER) pnhhsThisStub;
						}
						else
						{
							pntoaThisAdapter->Original_ReceivePacketHandler = NULL;
							pntoaThisAdapter->Stub_ReceivePacketHandler = NULL;
						}

						// hook transfer data complete handler in miniport block structure.

						if ( bHook && pmbMiniport && CanHook( pmbMiniport->TDCompleteHandler ) )
						{
							// save the pointer to the miniport block for restoring purposes later.

							pntoaThisAdapter->pmbMiniportPtr = pmbMiniport;

							// hook the function.

							pnhhsThisStub = & pnhhsStubsTemp[ dwStubsPos ++ ];
							FillStubCode( pnhhsThisStub, dwThisAdapterOrd, TDCOMPLETEHANDLER_FNID );
							pntoaThisAdapter->Original_TDCompleteHandler = pmbMiniport->TDCompleteHandler;
							pntoaThisAdapter->Stub_TDCompleteHandler = pnhhsThisStub;
							pmbMiniport->TDCompleteHandler = (NDIS_M_TD_COMPLETE_HANDLER) pnhhsThisStub;
						}
						else
						{
							// we will not hook this function.

							pntoaThisAdapter->Original_TDCompleteHandler = NULL;
							pntoaThisAdapter->Stub_TDCompleteHandler = NULL;
						}

					} while( pobAdapter = GetNextOpen( pobAdapter ) );
				}
			}
		}
		__except ( EXCEPTION_EXECUTE_HANDLER )
		{
			bFinalize = FALSE;
		}

		// We can continue only if there was no exception during the Hooking Operation.

		if ( bFinalize )
		{
			// Allocate the final list of Protocols and Adapters.

			dwProtocolListSize = sizeof( NT_PROTOCOL_LIST ) +
				ntplHeader.dwProtocolsNum * sizeof( NT_PROTOCOL ) +
				ntplHeader.dwOpenAdaptersNum * sizeof( NT_OPEN_ADAPTER );

			if ( dwProtocolListSize > PROTOCOLLIST_BUFFER_MAXSIZE )
				return STATUS_UNSUCCESSFUL;

			pplProtocolListTemp = (PNT_PROTOCOL_LIST) ExAllocatePool( NonPagedPool, dwProtocolListSize );
			if ( pplProtocolListTemp == NULL )
				return STATUS_UNSUCCESSFUL;

			pbBytePtr = (BYTE*) pplProtocolListTemp;

			// Copy the contents of the list.

			memcpy( pbBytePtr, & ntplHeader, sizeof( ntplHeader ) );
			pbBytePtr += sizeof( ntplHeader );

			memcpy( pbBytePtr, pntpProtocols, ntplHeader.dwProtocolsNum * sizeof( NT_PROTOCOL ) );
			g_pntpProtocols = (PNT_PROTOCOL) pbBytePtr; // <------------- !!! UPDATE THE GLOBAL POINTER HERE !!!
			pbBytePtr += ntplHeader.dwProtocolsNum * sizeof( NT_PROTOCOL );

			memcpy( pbBytePtr, pntoaAdapters, ntplHeader.dwOpenAdaptersNum * sizeof( NT_OPEN_ADAPTER ) );
			g_pntoaAdapters = (PNT_OPEN_ADAPTER) pbBytePtr; // <------------- !!! UPDATE THE GLOBAL POINTER HERE !!!

			// We can consider the operation as completed.

			* ppplProtocolList = pplProtocolListTemp;
			* ppnhhsStubs = pnhhsStubsTemp;
			* pdwStubsNum = dwStubsPos;

			nsReturnValue = STATUS_SUCCESS;
		}
	}
	__finally
	{
		// Eventually compensate an Unterminated Operation.

		if ( nsReturnValue != STATUS_SUCCESS && pntpProtocols && pntoaAdapters )
			UnhookInstalledProtocols( pntpProtocols, ntplHeader.dwProtocolsNum, pntoaAdapters, ntplHeader.dwOpenAdaptersNum );

		// Free the allocated memory.

		if ( pntpProtocols )
			ExFreePool( pntpProtocols );
		if ( pntoaAdapters )
			ExFreePool( pntoaAdapters );

		if ( pplProtocolListTemp && nsReturnValue != STATUS_SUCCESS )
		{
			ExFreePool( pplProtocolListTemp );
			* ppplProtocolList = NULL;
		}

		if ( pnhhsStubsTemp && nsReturnValue != STATUS_SUCCESS )
		{
			ExFreePool( pnhhsStubsTemp );
			* ppnhhsStubs = NULL;
		}
	}

	// Return to the caller.

	return nsReturnValue;
}

//===============================================
// UnhookInstalledProtocols Function Definition.
//===============================================

static void RestoreHandlerPointerSecure( PVOID pvDest, PVOID pvOriginal, PVOID pvSecureCheck )
{
	// Restore the Original Value only if we found the previous one that we have placed in the Hooking Phase.

	__try
	{
		if ( * (DWORD*) pvDest == (DWORD) pvSecureCheck )
			* (DWORD*) pvDest = (DWORD) pvOriginal;
	}
	__except ( EXCEPTION_EXECUTE_HANDLER )
	{
	}
}

NTSTATUS UnhookInstalledProtocols( PNT_PROTOCOL pntpProtocols, DWORD dwProtocolsSize, PNT_OPEN_ADAPTER pntoaAdapters, DWORD dwAdaptersSize )
{
	DWORD				i;
	PNT_OPEN_ADAPTER	pntoaThisAdapter;

	// Restore the original Handler Pointers.

	for ( i=0; i<dwAdaptersSize; i++ )
	{
		pntoaThisAdapter = & pntoaAdapters[ i ];

		// Restore the Original Pointers.

		if ( pntoaThisAdapter->Original_SendHandler )
			RestoreHandlerPointerSecure( & pntoaThisAdapter->pobBlockPtr->SendHandler, pntoaThisAdapter->Original_SendHandler, pntoaThisAdapter->Stub_SendHandler );
		if ( pntoaThisAdapter->Original_ReceiveHandler )
			RestoreHandlerPointerSecure( & pntoaThisAdapter->pobBlockPtr->ReceiveHandler, pntoaThisAdapter->Original_ReceiveHandler, pntoaThisAdapter->Stub_ReceiveHandler );
		if ( pntoaThisAdapter->Original_PostNt31ReceiveHandler )
			RestoreHandlerPointerSecure( & pntoaThisAdapter->pobBlockPtr->PostNt31ReceiveHandler, pntoaThisAdapter->Original_PostNt31ReceiveHandler, pntoaThisAdapter->Stub_PostNt31ReceiveHandler );
		if ( pntoaThisAdapter->Original_SendPacketsHandler )
			RestoreHandlerPointerSecure( & pntoaThisAdapter->pobBlockPtr->SendPacketsHandler, pntoaThisAdapter->Original_SendPacketsHandler, pntoaThisAdapter->Stub_SendPacketsHandler );
		if ( pntoaThisAdapter->Original_TransferDataHandler )
			RestoreHandlerPointerSecure( & pntoaThisAdapter->pobBlockPtr->TransferDataHandler, pntoaThisAdapter->Original_TransferDataHandler, pntoaThisAdapter->Stub_TransferDataHandler );
		if ( pntoaThisAdapter->Original_ReceivePacketHandler )
			RestoreHandlerPointerSecure( & pntoaThisAdapter->pobBlockPtr->ReceivePacketHandler, pntoaThisAdapter->Original_ReceivePacketHandler, pntoaThisAdapter->Stub_ReceivePacketHandler );
		if ( pntoaThisAdapter->Original_TDCompleteHandler && pntoaThisAdapter->pmbMiniportPtr )
			RestoreHandlerPointerSecure( & pntoaThisAdapter->pmbMiniportPtr->TDCompleteHandler, pntoaThisAdapter->Original_TDCompleteHandler, pntoaThisAdapter->Stub_TDCompleteHandler );
	}

	// Return to the caller.

	return STATUS_SUCCESS;
}
