
#include "NdisHook.h"
#include "vpcknt.h"

//==================================
// Boilerplate KM device functions.
//==================================

VOID VPCKernelNtDriverUnload( IN PDRIVER_OBJECT DriverObject );
NTSTATUS VPCKernelNtDispatchCreate( IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp );
NTSTATUS VPCKernelNtDispatchClose( IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp );
NTSTATUS VPCKernelNtDispatchIoctl( IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp );

//=================================================
// Private storage for us to squirrel things away.
//=================================================

// *** Device Object Global Pointer. ***

PDEVICE_OBJECT		g_pDeviceObject = NULL;

//==================================
// DriverEntry Function Definition.
//==================================

NTSTATUS DriverEntry( IN PDRIVER_OBJECT DriverObject, IN PUNICODE_STRING RegistryPath )
{
	NTSTATUS        ntStatus;
	UNICODE_STRING  uszDriverString;
	UNICODE_STRING  uszDeviceString;

	PDEVICE_OBJECT    pDeviceObject;
	PDEVICE_EXTENSION extension;

	// Point uszDriverString at the driver name.

	RtlInitUnicodeString( & uszDriverString, L"\\Device\\" SYSDRIVER_NAME_WIDE );

	// Create and initialize device object.

	ntStatus = IoCreateDevice( DriverObject,
		sizeof( DEVICE_EXTENSION ),
		& uszDriverString,
		FILE_DEVICE_UNKNOWN,
		0,
		FALSE,
		& pDeviceObject );

	if ( ntStatus != STATUS_SUCCESS )
	{
		// Return to the Caller.

		return ntStatus;
	}

	// Assign extension variable.

	extension = pDeviceObject->DeviceExtension;

	// Point uszDeviceString at the device name.

	RtlInitUnicodeString( & uszDeviceString, L"\\DosDevices\\" SYSDRIVER_NAME_WIDE );

	// Create symbolic link to the user-visible name.

	ntStatus = IoCreateSymbolicLink( & uszDeviceString, & uszDriverString );

	if ( ntStatus != STATUS_SUCCESS )
	{
		// Delete device object if not successful.

		IoDeleteDevice( pDeviceObject );

		return ntStatus;
	}

	// Assign global pointer to the device object.

	g_pDeviceObject = pDeviceObject;

	// Load structure to point to IRP handlers.

	DriverObject->DriverUnload                           = VPCKernelNtDriverUnload;
	DriverObject->MajorFunction[ IRP_MJ_CREATE ]         = VPCKernelNtDispatchCreate;
	DriverObject->MajorFunction[ IRP_MJ_CLOSE ]          = VPCKernelNtDispatchClose;
	DriverObject->MajorFunction[ IRP_MJ_DEVICE_CONTROL ] = VPCKernelNtDispatchIoctl;

	// Initialize WasCreated and its mutex.

	extension->WasCreated = 0;
	ExInitializeFastMutex( & extension->WasCreatedFastMutex );

	// Initialize InterceptNDIS.

	extension->InterceptNDIS = 0;

	// Initialize the NDIS State variables.

	extension->NdisProtocolList = NULL;
	extension->NdisStubsList = NULL;
	extension->NdisStubsNum = 0;

	// Initialize the OaList stuff.

	extension->OaList = NULL;
	extension->OaListItems = 0;
	NdisAllocateSpinLock( & extension->OaListSpinLock );

	// Initialize the packets buffer.

	extension->PacketsBuff = NULL;
	extension->PacketsBuffMaxItems = 0;

	extension->PacketsBuffStart = -1;
	extension->PacketsBuffEnd = 0;

	extension->PacketsLost = 0;

	extension->PacketsReadyEvent = NULL;

	NdisAllocateSpinLock( & extension->PacketsBuffSpinLock );

	// Initialize the other res...

	NdisAllocateSpinLock( & extension->ReceiveWorkItemSpinLock );
	extension->ReceiveWorkItemIdCounter = 0;

	NdisAllocateSpinLock( & extension->PacketSerialCounterSpinLock );
	extension->PacketSerialCounter = 0;

	// Initialize the lt stuff.

	extension->dwLtLoopbackAdapterIpAddress = 0;
	extension->dwLtTrickNatIpAddress = 0;

	// Return Success.

	return STATUS_SUCCESS;
}

//====================
// cleanup functions.
//====================

static VOID ReleaseOaList ()
{
	PDEVICE_EXTENSION	extension = g_pDeviceObject->DeviceExtension;

	NdisAcquireSpinLock( & extension->OaListSpinLock );

		if ( extension->OaList )
		{
			// free.

			ExFreePool( extension->OaList );

			extension->OaList = NULL;
			extension->OaListItems = 0;
		}

	NdisReleaseSpinLock( & extension->OaListSpinLock );
}

static VOID ReleasePackets ()
{
	PDEVICE_EXTENSION	extension = g_pDeviceObject->DeviceExtension;

	NdisAcquireSpinLock( & extension->PacketsBuffSpinLock );

		if ( extension->PacketsBuff )
		{
			if ( extension->PacketsBuffStart != -1 )
			{
				DWORD		dw;

				DWORD		dwMax = extension->PacketsBuffEnd;
				if ( extension->PacketsBuffEnd <= extension->PacketsBuffStart )
					dwMax = extension->PacketsBuffMaxItems;

				for ( dw=extension->PacketsBuffStart; dw < dwMax; dw ++ )
				{
					NEXT_PACKET*		pThis = & extension->PacketsBuff[ dw ];

					// free the packet.
					ExFreePool( pThis->pbData );
				}

				if ( dwMax != extension->PacketsBuffEnd )
				{
					for( dw=0; dw<extension->PacketsBuffEnd; dw ++ )
					{
						NEXT_PACKET*		pThis = & extension->PacketsBuff[ dw ];

						// free the packet.
						ExFreePool( pThis->pbData );
					}
				}
			}

			ExFreePool( extension->PacketsBuff );

			extension->PacketsBuff = NULL;
			extension->PacketsBuffMaxItems = 0;

			extension->PacketsBuffStart = -1;
			extension->PacketsBuffEnd = 0;
		}

		extension->PacketsLost = 0;

		if ( extension->PacketsReadyEvent )
		{
			ObDereferenceObject( extension->PacketsReadyEvent );
			extension->PacketsReadyEvent = NULL;
		}

	NdisReleaseSpinLock( & extension->PacketsBuffSpinLock );
}

static VOID ReleaseReceiveWorkItems ()
{
	PDEVICE_EXTENSION	extension = g_pDeviceObject->DeviceExtension;
	int					i, c;
	NT_PROTOCOL*		pBase = NULL;
	NT_PROTOCOL*		pThis = NULL;

	if ( extension->NdisProtocolList == NULL ||
		extension->NdisProtocolList->dwProtocolsNum == 0 )
	{
		return;
	}

	pBase = (NT_PROTOCOL*) ( (BYTE*) extension->NdisProtocolList + sizeof( NT_PROTOCOL_LIST ) );

	NdisAcquireSpinLock( & extension->ReceiveWorkItemSpinLock );

		c = extension->NdisProtocolList->dwProtocolsNum;
		for ( i=0; i<c; i ++ )
		{
			pThis = & pBase[ i ];

			if ( pThis->pbWorkItemHeader )
			{
				// free...

				ExFreePool( pThis->pbWorkItemHeader );

				pThis->pbWorkItemHeader = NULL;
				pThis->uiWorkItemHeaderSize = 0;
			}
		}

	NdisReleaseSpinLock( & extension->ReceiveWorkItemSpinLock );
}

//===================================
// DriverUnload Function Definition.
//===================================

VOID VPCKernelNtDriverUnload( IN PDRIVER_OBJECT DriverObject )
{
	PDEVICE_EXTENSION	extension = g_pDeviceObject->DeviceExtension;
	UNICODE_STRING		uszDeviceString;

	// Free resources.

	VPCKernelNtDispatchClose( g_pDeviceObject, NULL );

	ReleaseReceiveWorkItems ();
	ReleaseOaList ();
	ReleasePackets ();

	NdisFreeSpinLock( & extension->ReceiveWorkItemSpinLock );
	NdisFreeSpinLock( & extension->OaListSpinLock );
	NdisFreeSpinLock( & extension->PacketsBuffSpinLock );
	NdisFreeSpinLock( & extension->PacketSerialCounterSpinLock );

	if ( extension->NdisStubsList )
	{
		ExFreePool( extension->NdisStubsList );
		extension->NdisStubsList = NULL;
	}

	if ( extension->NdisProtocolList )
	{
		ExFreePool( extension->NdisProtocolList );
		extension->NdisProtocolList = NULL;
	}

	if ( extension->OaList )
	{
		ExFreePool( extension->OaList );
		extension->OaList = NULL;
	}

	if ( extension->PacketsBuff )
	{
		ExFreePool( extension->PacketsBuff );
		extension->PacketsBuff = NULL;
	}

	// Do several Clean-Up operations before returning control.

	RtlInitUnicodeString( & uszDeviceString, L"\\DosDevices\\" SYSDRIVER_NAME_WIDE );

	IoDeleteSymbolicLink( & uszDeviceString );
	IoDeleteDevice( g_pDeviceObject );

	g_pDeviceObject = NULL;

	// Return to the Caller.

	return;
}

//================================================
// VPCKernelNtDispatchCreate Function Definition.
//================================================

NTSTATUS VPCKernelNtDispatchCreate( IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp )
{
	PDEVICE_EXTENSION extension = DeviceObject->DeviceExtension;
	ULONG             WasCreated;
	NTSTATUS          ntStatus;

	// Allow only one creation of the file object

	ExAcquireFastMutex( & extension->WasCreatedFastMutex );
	{
		WasCreated = extension->WasCreated;
		if ( extension->WasCreated == 0 )
			extension->WasCreated = 1;
	}
	ExReleaseFastMutex( & extension->WasCreatedFastMutex );

	if ( WasCreated != 0 )
		ntStatus = STATUS_UNSUCCESSFUL;
	else
		ntStatus = STATUS_SUCCESS;

	// Return to the caller

	Irp->IoStatus.Status = ntStatus;
    Irp->IoStatus.Information = 0;

    IoCompleteRequest( Irp, IO_NO_INCREMENT );

    return ntStatus;
}

//===============================================
// VPCKernelNtDispatchClose Function Definition.
//===============================================

NTSTATUS VPCKernelNtDispatchClose( IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp )
{
	PDEVICE_EXTENSION		extension = DeviceObject->DeviceExtension;
	ULONG					bDoCleanUp = 0;

	// Remember that the client is closing the device

	ExAcquireFastMutex( & extension->WasCreatedFastMutex );

		if ( extension->WasCreated )
			bDoCleanUp = 1;
		extension->WasCreated = 0;

	ExReleaseFastMutex( & extension->WasCreatedFastMutex );

	// Check whether we have to do Clean-Up.

	if ( bDoCleanUp )
	{
		LARGE_INTEGER			liWaitElapse;

		liWaitElapse = RtlConvertLongToLargeInteger( - 5000000 ); // Half a second.

		// ### ### ### ### ### NDIS CLEAN-UP PROCEDURE ### ### ### ### ###

		// Erase the list of subscribed open adapters.

		ReleaseOaList ();

		// Stop Intercepting.

		extension->InterceptNDIS = 0;

		KeDelayExecutionThread( UserMode, FALSE, & liWaitElapse );

		// Unhook NDIS Handler functions.

		if ( extension->NdisProtocolList )
			UnhookInstalledProtocols(
				(PNT_PROTOCOL) ( (BYTE*) extension->NdisProtocolList + sizeof( NT_PROTOCOL_LIST ) ),
				extension->NdisProtocolList->dwProtocolsNum,
				(PNT_OPEN_ADAPTER) ( (BYTE*) extension->NdisProtocolList + sizeof( NT_PROTOCOL_LIST ) + extension->NdisProtocolList->dwProtocolsNum * sizeof( NT_PROTOCOL ) ),
				extension->NdisProtocolList->dwOpenAdaptersNum );

		KeDelayExecutionThread( UserMode, FALSE, & liWaitElapse );

		// Free the memory.

		ReleaseReceiveWorkItems ();
		ReleasePackets ();

		if ( extension->NdisProtocolList )
		{
			ExFreePool( extension->NdisProtocolList );
			extension->NdisProtocolList = NULL;
		}

		if ( extension->NdisStubsList )
		{
			ExFreePool( extension->NdisStubsList );
			extension->NdisStubsList = NULL;
		}

		extension->NdisStubsNum = 0;
	}

	// Return to the caller

	if ( Irp )
	{
		Irp->IoStatus.Status = STATUS_SUCCESS;
		Irp->IoStatus.Information = 0;

		IoCompleteRequest( Irp, IO_NO_INCREMENT );
	}

    return STATUS_SUCCESS;
}

//===============================================
// VPCKernelNtDispatchIoctl Function Definition.
//===============================================

NTSTATUS VPCKernelNtDispatchIoctl( IN PDEVICE_OBJECT DeviceObject, IN PIRP Irp )
{
    NTSTATUS              ntStatus = STATUS_UNSUCCESSFUL;
    PIO_STACK_LOCATION    irpStack  = IoGetCurrentIrpStackLocation(Irp);
    PDEVICE_EXTENSION     extension = DeviceObject->DeviceExtension;

	DWORD*                pl;

	// IOCTL Processor Switch.

	switch( irpStack->Parameters.DeviceIoControl.IoControlCode )
	{
		//-----------------------------------------------------------------------------------

		case IOCTL_VPCKNT_GET_VERSION:
			{
				// Return the information required.

				if ( irpStack->Parameters.DeviceIoControl.OutputBufferLength >= sizeof( DWORD ) )
				{
					// Copy the information in the system buffer.

					pl = Irp->AssociatedIrp.SystemBuffer;
					*pl = ( DWORD ) _VPCKnt_DRIVER_VERSION_;

					// Everything went ok.

					ntStatus = STATUS_SUCCESS;
				}
			}
			break;

		//-----------------------------------------------------------------------------------

		case IOCTL_VPCKNT_INITIALIZE_HOOK:
			{
				// === Initialize the System Wide hooks. ===

				if ( irpStack->Parameters.DeviceIoControl.InputBufferLength == INITIALIZE_HOOK_INPUT_SIZE &&
					irpStack->Parameters.DeviceIoControl.OutputBufferLength == INITIALIZE_HOOK_OUTPUT_MAXSIZE )
				{
					INITIALIZE_HOOK_INPUT		ihiHookInput;
					INITIALIZE_HOOK_OUTPUT		ihoHookOutput;
					NTSTATUS					nsHookRes = STATUS_UNSUCCESSFUL;
					PNT_PROTOCOL_LIST			pplProtocolList;
					PNDISHOOK_HANDLER_STUB		pnhhsStubsList;
					DWORD						dwStubsNum;
					NDIS_HANDLE					hFakeProtocolHandle = NULL;
					NTSTATUS					nsRegRes = STATUS_UNSUCCESSFUL;
					BYTE*						pbOutputBufferPayload;

					ihiHookInput = * (INITIALIZE_HOOK_INPUT*) Irp->AssociatedIrp.SystemBuffer;
					memset( & ihoHookOutput, 0, sizeof( ihoHookOutput ) );

					ntStatus = STATUS_SUCCESS; // ---> We return Detailed Informations about errors in the Output Structure.

					// ### ### ### ### ### ALLOCATE PACKETS BUFFER ### ### ### ### ###

					extension->PacketsBuff = (NEXT_PACKET*) ExAllocatePool( NonPagedPool,
						ihiHookInput.dwPacketsBuffMaxItems * sizeof( NEXT_PACKET ) );

					if ( extension->PacketsBuff )
					{
						extension->PacketsBuffMaxItems = ihiHookInput.dwPacketsBuffMaxItems;
						memset( extension->PacketsBuff, 0, ihiHookInput.dwPacketsBuffMaxItems * sizeof( NEXT_PACKET ) );
					}

					if ( ihiHookInput.hNotificationEvent )
					{
						NTSTATUS		ntEvRes = ObReferenceObjectByHandle( ihiHookInput.hNotificationEvent,
							0, (POBJECT_TYPE) NULL, UserMode, (PVOID) & extension->PacketsReadyEvent, (POBJECT_HANDLE_INFORMATION) NULL);

						if ( ntEvRes != STATUS_SUCCESS )
							extension->PacketsReadyEvent = NULL;
					}

					// ### ### ### ### ### HOOK NDIS ### ### ### ### ###

					// Start-Up.

					__try
					{
						nsRegRes = RegisterFakeProtocol( & hFakeProtocolHandle, SYSDRIVER_NAME_ANSI );

						if ( nsRegRes == STATUS_SUCCESS )
						{
							nsHookRes = HookInstalledProtocols(
								& pplProtocolList, & pnhhsStubsList, & dwStubsNum,
								hFakeProtocolHandle );
						}
					}
					__except ( EXCEPTION_EXECUTE_HANDLER )
					{
					}

					// Clean-Up.

					if ( nsRegRes == STATUS_SUCCESS && hFakeProtocolHandle )
						DeregisterFakeProtocol( hFakeProtocolHandle );

					// ### ### ### ### ### OUTPUT BUFFER ### ### ### ### ###

					// Set the Payload pointer.

					pbOutputBufferPayload = (BYTE*) Irp->AssociatedIrp.SystemBuffer + sizeof( ihoHookOutput );

					// Inform about the Ndis Hook results.

					if ( nsHookRes == STATUS_SUCCESS ) // NDIS Hook Ok.
					{
						// Info.

						ihoHookOutput.bNdisHookSucceeded = 1;
						ihoHookOutput.dwProtocolListBufferSize = sizeof( NT_PROTOCOL_LIST ) +
							pplProtocolList->dwProtocolsNum * sizeof( NT_PROTOCOL ) +
							pplProtocolList->dwOpenAdaptersNum * sizeof( NT_OPEN_ADAPTER );

						// Data.

						memcpy( pbOutputBufferPayload, pplProtocolList, ihoHookOutput.dwProtocolListBufferSize );
						pbOutputBufferPayload += ihoHookOutput.dwProtocolListBufferSize;
					}
					else // NDIS Hook Failed.
					{
						// Info.

						ihoHookOutput.bNdisHookSucceeded = 0;
						ihoHookOutput.dwProtocolListBufferSize = 0;
					}

					// Copy the Header structure.

					* (INITIALIZE_HOOK_OUTPUT*) Irp->AssociatedIrp.SystemBuffer = ihoHookOutput;

					// ### ### ### ### ### DEVICE EXTENSION ### ### ### ### ###

					if ( nsHookRes == STATUS_SUCCESS ) // NDIS Hook Ok.
					{
						extension->NdisProtocolList = pplProtocolList;
						extension->NdisStubsList = pnhhsStubsList;
						extension->NdisStubsNum = dwStubsNum;
					}
					else // NDIS Hook Failed.
					{
						extension->NdisProtocolList = NULL;
						extension->NdisStubsList = NULL;
						extension->NdisStubsNum = 0;
					}

					// ### ### ### ### ### HOOK ACTIVATION ### ### ### ### ###

					if ( nsHookRes == STATUS_SUCCESS )
					{
						extension->dwLtLoopbackAdapterIpAddress = ihiHookInput.dwLtLoopbackAdapterIpAddress;
						extension->dwLtTrickNatIpAddress = ihiHookInput.dwLtTrickNatIpAddress;
					}
					else
					{
						extension->dwLtLoopbackAdapterIpAddress = 0;
						extension->dwLtTrickNatIpAddress = 0;
					}

					extension->InterceptNDIS = 1;
				}
			}
			break;

		//-----------------------------------------------------------------------------------

		case IOCTL_VPCKNT_SUBMIT_OALIST:
			{
				// === copy the list ===

				if (
						( irpStack->Parameters.DeviceIoControl.InputBufferLength % sizeof( OALIST_ITEM ) ) == 0
					)
				{
					OALIST_ITEM*			pList = (OALIST_ITEM*) Irp->AssociatedIrp.SystemBuffer;
					DWORD					dwListItems = irpStack->Parameters.DeviceIoControl.InputBufferLength / sizeof( OALIST_ITEM );

					// return success.

					ntStatus = STATUS_SUCCESS;

					//
					// make a copy of the list...
					//

					NdisAcquireSpinLock( & extension->OaListSpinLock );

						// free the prev.

						if ( extension->OaList )
						{
							// free.

							ExFreePool( extension->OaList );

							extension->OaList = NULL;
							extension->OaListItems = 0;
						}

						// allocate the new.

						if ( pList )
						{
							extension->OaList = ExAllocatePool( NonPagedPool, dwListItems * sizeof( OALIST_ITEM ) );
							if ( extension->OaList )
							{
								memcpy( extension->OaList, pList, dwListItems * sizeof( OALIST_ITEM ) );
								extension->OaListItems = dwListItems;
							}
						}
						else
						{
							extension->OaList = NULL;
							extension->OaListItems = 0;
						}

					NdisReleaseSpinLock( & extension->OaListSpinLock );
				}
			}
			break;

		//-----------------------------------------------------------------------------------

		case IOCTL_VPCKNT_GET_NEXT_PACKET:
			{
				NEXT_PACKET		npPacket;
				HOOK_STATS		hsStats;

				BOOL			bRetrRes;

				// === return the info. ===

				ntStatus = STATUS_UNSUCCESSFUL;

				bRetrRes = RetrievePacket( & npPacket, & hsStats );
				if ( bRetrRes )
				{
					BYTE*			pbOut = (BYTE*) Irp->AssociatedIrp.SystemBuffer;
					DWORD			size = sizeof( HOOK_STATS ) +
						sizeof( NEXT_PACKET ) + npPacket.dwDataLength;

					if ( irpStack->Parameters.DeviceIoControl.OutputBufferLength >= size )
					{
						//
						// copy the structures.
						//

						* (HOOK_STATS*) pbOut = hsStats;
						pbOut += sizeof( HOOK_STATS );

						* (NEXT_PACKET*) pbOut = npPacket;
						( (NEXT_PACKET*) pbOut )->pbData = NULL;
						pbOut += sizeof( NEXT_PACKET );

						memcpy( pbOut, npPacket.pbData, npPacket.dwDataLength );

						//
						// set ret val.
						//

						ntStatus = STATUS_SUCCESS;

						irpStack->Parameters.DeviceIoControl.OutputBufferLength = size;
					}

					ExFreePool( npPacket.pbData );
				}
			}
			break;

		//-----------------------------------------------------------------------------------

        default:
            break;

		//-----------------------------------------------------------------------------------
	}

	// Set the Return Value.

    Irp->IoStatus.Status = ntStatus;
    
    // Set # of bytes to copy back to user-mode.

    if ( ntStatus == STATUS_SUCCESS )
        Irp->IoStatus.Information = irpStack->Parameters.DeviceIoControl.OutputBufferLength;
    else
        Irp->IoStatus.Information = 0;

	// Return to the Caller.

    IoCompleteRequest( Irp, IO_NO_INCREMENT );
    return ntStatus;
}
