%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Set variable values
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
v = .10;
w = .50;
x = .10;
y = .10;
z = .20;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Constant Values
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
mac_oh = 40; % MAC Layer overhead in (Bytes)
phy_oh = 52.5; % PHY Layer overhead (Bytes)
kb = 1.38 * 10^-23; % Boltzmanns constant (J/K)
T = 290; % System noise temperatue (K)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Minimum and Maximum values
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
power_max = 24; % Maximum Power (dBm)
power_min = -8; % Minimum Power (dBm)
bandwidth_max = 34; % Maximum bandwidth (MHz)
115bandwidth_min = 2; % Minimum bandwidth (MHz)
mod_max = 256; % Maximum modulation index.
mod_min = 2;
symbol_rate_max = 1000000; % Maximum symbol rate (Symbols / second)
symbol_rate_min = 125000;
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Transmission paramet er ranges
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
power = [power_min:2:power_max];
bandwidth = [bandwidth_min:2:bandwidth_max];
tdd = [1:10:100];
mod_index = [ 2 4 16 64 256 ];
symbol_rate = [125000 250000 500000 625000 750000 1000000];
frame_length = [100 : 100 : 1500 ];
coding_rate = [1 1/2 3/4 5/6];
mod_type = {qam,psk,pam};
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Environmental parameter ranges
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
path_loss = [87:1:92];
noise_power = [-117:1:-112];
% Total number of parameter combinations
combinations = length(power) * length(bandwidth) ...
* length(tdd) * length(mod_index) * ...
length(symbol_rate) * length(frame_length) ...
* length(mod_type) * length(coding_rate) * ...
length(path_loss) * length(noise_power)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Fitness Function
%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Initialize the indexes that hold the position in the array
mod_index_index = 1;
tdd_index = 1;
symbol_rate_index = 1;
116power_index = 1;
bandwidth_index = 1;
mod_type_index = 1;
coding_rate_index = 1;
frame_length_index = 1;
snr_index = 1;
np_index = 1;
pl_index = 1;
fitness_array = [ ];
% Go through all possible combinations of transmission parameters and
% calculate the fitness. These fitness values are held in a 10-dimensional
% array indexed by the parameter indexes defined above.
for p = power
for b = bandwidth
for m = mod_index
for pl = path_loss
for sr = symbol_rate
for np = noise_power
for cr = coding_rate
for mt = mod_type
for t = tdd
for f = frame_length
%% Populate the fitness array with
%% outputs from the fitness
%% function
fitness_array(np_index,pl_index,power_index,
bandwidth_index,tdd_index,mod_index_index,
symbol_rate_index,frame_length_index,
coding_rate_index,mod_type_index) = ...
fitness_score(np,pl,p,b,t,m,sr,f,cr,char(mt),v,w,x,y,z,
power_max,power_min,symbol_rate_max,symbol_rate_min,
bandwidth_max,bandwidth_min,mod_max,mod_min);
frame_length_index = frame_length_index + 1;
end
tdd_index = tdd_index + 1;
frame_length_index = 1;
117end
tdd_index = 1;
mod_type_index = mod_type_index + 1;
end
mod_type_index = 1;
coding_rate_index = coding_rate_index + 1;
end
coding_rate_index = 1;
np_index =np_index + 1;
end
np_index = 1;
symbol_rate_index = symbol_rate_index + 1;
end
symbol_rate_index = 1;
pl_index = pl_index + 1;
end
pl_index = 1;
mod_index_index = mod_index_index + 1;
end
mod_index_index = 1;
bandwidth_index = bandwidth_index + 1;
end
power_index = power_index + 1;
bandwidth_index = 1;
end
% Determine the maximum,minimum, and range of fitness
% for each environment.
for i = 1:1:6
for j = 1:1:6
A = fitness_array(i,j,:,:,:,:,:,:,:,:);
max_fitness(i,j) = max(A(:));
min_fitness(i,j) = min(A(:));
range_fitness(i,j) = max(A(:)) - min(A(:));
end
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Open CLIPS rules file for writing
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
118filename = sprintf(rules/Cognitive_Rules_emerg_1.clp);
fid = fopen(filename, w);
if (fid == -1)
error(write_file: cannot open file for writing);
end
% Loop through each environment and create a CLIPS rule for the
% optimal fitness found.
for i = 1:1:6
for j = 1:1:6
rule_count = rule_count + 1;
% Find the max fitness and search for the corresponding
% transmission parameters in the array.
A = fitness_array(i,j,:,:,:,:,:,:,:,:);
[ o k ] = max(A(:));
[np_index pl_index power_index bandwidth_index
tdd_index mod_index_index symbol_rate_index
frame_length_index coding_rate_index mod_type_index]
= ind2sub(size(A),k);
% Translate the index found for the max fitness
% into real parameter values.
p = power(power_index);
b = bandwidth(bandwidth_index);
t = tdd(tdd_index);
mi = mod_index(mod_index_index);
sr = symbol_rate(symbol_rate_index);
f = frame_length(frame_length_index);
cr = coding_rate(coding_rate_index);
mt = char(mod_type(mod_type_index));
%% Print rules to the CLIPS rules file
fprintf(fid, (defrule cognitive_rule_%i\n,rule_count);
fprintf(fid,  (noise_power %i)\n,-118+i);
fprintf(fid,  (path_loss %i)\n,86+j);
fprintf(fid,  (channel_number ?channel_num)\n);
fprintf(fid,  (scenario emerg_mode)\n);
119fprintf(fid, =>\n);
fprintf(fid,  (assert (channel ?channel_num %2.0f
%2.0f %3.0f %s %4.0f %1.2f %7.0f %3.0f)) )\n\n\n,
p,b,mi,mt,f,cr,sr,t);
end
end