

#pragma once

#include "node.h"
#include "Y_TAB.h"
#include <process.h> 
#include <windows.h>

#include <stdio.h>
#include <stdlib.h>
#include <tchar.h>
#include <conio.h>
#include <iostream>
#include <math.h>
#include <utility>
#include <map>
#include <string>
#include <vector>

using namespace std;

#define MIN_UTILITY 0
#define MAX_UTILITY 100
#define MAX_UTILITY_BUFFER 20
#define CAPRI_SERVER_INTERVAL 250
#define MAX_APPS 5
#define STEP_MAX_DATA_RATE 2000000

#define ANY_CHARACTER "*"
#define ANY 0
#define MAX_SPEC 1028
#define CAPRI_PORT 4002
#define MAX_ALA_LEVELS 10

typedef struct _FlowID_t {
	unsigned int protocol;   //UDP 17, TCP 6, based on ws2def.h and oihnet.h
	unsigned long srcAddr;
	unsigned long dstAddr;
	unsigned short dstPort;
	unsigned short srcPort;
} FlowID;

typedef struct _FlowStatistics
{
	FlowID	id;
	ULONG	TxNumBytes;
	ULONG	RxNumBytes;
	ULONG	TxNumBytesLink;		// the data sent over the current link
	ULONG	RxNumBytesLink;		// the data received through the current link
	ULONG	TxBitRate;
	ULONG	RxBitRate;
	INT		latency;
	INT		jitter;
	LONG	receivedUDPBytes;
	LONG	lostUDPBytes;
	INT		appType;
	INT		priority;
} FlowStatistics;

typedef struct CapriAttrDescr_t {
	int id;
	int action;
	int attribute;
	float data;
} CapriAttrDescr;

typedef enum _APP_TYPE {
	CAPRI_APP_VIDEO_STREAMING = 101,
	CAPRI_APP_TCP_DOWNLOAD_2,
	CAPRI_APP_TCP_DOWNLOAD_1,
	CAPRI_APP_TCP_DOWNLOAD,
	CAPRI_APP_UDP_DOWNLOAD,
	CAPRI_APP_AUDIO,
	CAPRI_APP_DOWNLOAD,
	CAPRI_APP_WWW,
	CAPRI_APP_TEST,
	CAPRI_APP_DEFAULT,
} APP_TYPE;

typedef map<int, utilityTree *> mapUtility;
typedef map<int, const FlowID> mapFlow;

typedef map<int, map<int, float > > mapAttribute;
typedef map<int, map<string, int> > mapUnknownConst; 

typedef map<int, string> mapString;
typedef map<int, SOCKET> mapApp;
typedef map<APP_TYPE, utilityTree *> mapTree;
typedef map<int, map<int, APP_TYPE>> mapRange;

typedef enum _PROTOCOL_APP_TYPE {
	PROTOCOL_APP_VIDEO_STREAMING = 101,
	PROTOCOL_APP_DOWNLOAD,
	PROTOCOL_APP_TEST,
	PROTOCOL_APP_IPERF,
} PROTOCOL_APP_TYPE;

typedef enum _CAPRI_ERROR_CODE {
	CAPRI_OK = 0,
	CAPRI_FAILED = -1,
	CAPRI_INDEX_NOT_FOUND = -2,
} CAPRI_ERROR_CODE;

typedef enum _PROTOCOL_TYPE {
	PROTOCOL_TCP = 201,
	PROTOCOL_UDP,
	PROTOCOL_UNDEFINED = 0,
} PROTOCOL_TYPE;

typedef enum _CONFIGURABILITY_TYPE {
	CONFIGURABILITY_STATIC = 301,
	CONFIGURABILITY_DYNAMIC,
	CONFIGURABILITY_TRANSIENT,
} CONFIGURABILITY_TYPE;

typedef enum _UTILITY_FLAG {
	DEFAULT_UTILITY_IF_INVALID = 0,
	INVALID_UTILITY = 111,
	VALID_UTILITY = 222,
} UTILITY_FLAG;
 
typedef enum _CAPRI_ATTRIBUTE {
	ATTR_MAX_DATA_RATE = 22,
	
} CAPRI_ATTRIBUTE;

typedef struct CapriRegister_t {
	FlowID id;
	APP_TYPE appType;
	char utilSpec[MAX_SPEC];
	char utilConst[MAX_SPEC];
} CapriRegister;

typedef struct CapriSocket_t {
	INT action;
	INT appType;
	union{
		CapriRegister cr;
		CapriAttrDescr ca;
		INT fid;
	};
	
} CapriSocket;

// evaluate utility with a given utility tree 
FLOAT evaluateByUtilityTree(const INT fid, utilityTree *p);

// normalize a given utility
FLOAT normalizeUtility(const INT fid, float util);

// register with a utility and const tree which can be retrieved from the table
INT registerUtilityTreeWithConfigurability(const FlowID *id, utilityTree *uTree, utilityTree *constTree, INT configurability, APP_TYPE appType);

// register utility with utility string
INT registerUtilityWithConfigurability(const FlowID *id, const char *utilSpec, const char *utilConst, INT configurability, APP_TYPE appType);

// get a default utility specification from a config file 
const char *getDefaultUtilitySpec(INT appType);

// get a default utility constraint specification from a config file 
const char *getDefaultConstraintSpec(INT appType);

// check if CAPRI is still running, otherwise kill all threads
bool capriIsRunning();

 unsigned int  WINAPI startServer(void *arg); 
INT startServerThread();
INT initWinSock();

// process an incoming data packet from socket.
void processSocketData(SOCKET remoteSocket, CapriSocket * cs);


// notify CRM when indices are removed from the utility table
void notifyRemovedIndex(vector<FlowID> removedFlowId);


