
#pragma once

#include "Capri.h"
#include <vector>

class ICapriNotifySink
{
public:
	// callback function for notifying CRM when configuration is reloaded
	virtual void CapriConfigReloaded(std::vector<FlowID> removedFlowId) = 0;
		
protected:
	virtual ICapriNotifySink::~ICapriNotifySink() = 0;

};

// initialize the CAPRI core and give a callback.
INT initCapri(char *confFile, USHORT portCapri, ICapriNotifySink *pCaprinNotification);

INT closeCapri();

// get a default application type by a given port
APP_TYPE getAppTypeByPort(const FlowID *id);

// get an application type according to an application port.
APP_TYPE getAppType(FlowID *pFlowID);

// evaluate utility by giving an index and measured parameters
FLOAT evaluate(const INT fid, FLOAT delay, FLOAT throughput, FLOAT packetloss, FLOAT jitter);

// evaluate utility by giving FlowID and measured parameters
FLOAT evaluateByFlowID(const FlowID *id, INT *fid, FLOAT delay, FLOAT throughput, FLOAT packetloss, FLOAT jitter);

// remove a given FlowID 
INT removeFlowIndex(const FlowID *id);

// to be called in SessionManager::onNewFlow when a new flow is detected by flowcontrol.cpp
INT onDetectedFlow(const FlowID *id);

// get a requested attribute through socket.
INT capriGetAttr(CapriAttrDescr *attr, INT aid);

// set a given attribute value through socket.
INT capriSetAttr(CapriAttrDescr *attr, INT aid);

// check a tuning decision has to be triggered.
bool checkTuneParam(const int fid, int appType, float desiredUtilDiffTh, float desiredUtilTh);

// tune application parameters throught socket.
INT tuneAppParam(int command, int action, FlowID *id, int level);

// restore tunable application parameters to initial pre-defined values.
INT restoreAppParam(FlowID *id);

// check is an application is capri-aware (its parameter is tunable).
INT isAppCapriAware(FlowID *id);
