

#ifndef CONFIGFILE_H
#define CONFIGFILE_H

#include <string>
#include <map>
#include <iostream>
#include <fstream>
#include <sstream>
#include "Capri.h"

using namespace std;

#define CAPRICONFIGFILE "C:\\capriconf.txt"
#define CONFIG_UPDATE_INTERVAL 3000
#define CONFIG_UPDATE_MAX_COUNT_TIMER 0  // nonstop


class ConfigFile {

protected:
	string lineSeparator; 
	string valueSeparator; 
	string Comment;
	map<string,map<string,string>> myData; 
	
	typedef map<string,string>::const_iterator map_it1d;
	typedef map<string,map<string,string>>::const_iterator map_it;

public:
	ConfigFile( string filename,
				string lineSep = "=",
	            string valueSep = ";",
				string comment = "%" );
	ConfigFile();
	~ConfigFile();
	
	template<class T> T read( string key, string key2 ) ; 
	template<class T> void readFromKey( T& var, const string& key ) const;
	
	template<class T> void add( string key, string key2, const T& value );
	void remove( const string& key1 , const string &key2);
	
	string getLineSeparator() const { return lineSeparator; }
	string setLineSeparator( const string& s )
		{ string old = lineSeparator;  lineSeparator = s;  return old; }  
	string getValueSeparator() const { return valueSeparator; }
	string setValueSeparator( const string& s )
		{ string old = valueSeparator;  valueSeparator = s;  return old; }
	friend ostream& operator<<( ostream& os, const ConfigFile& cf );
	friend istream& operator>>( istream& is, ConfigFile& cf );

	void registerApplication(APP_TYPE appType);

	void registerPortRange(APP_TYPE appType, string portRange);

	void registerAllDefaultApplication();
	void registerAllApplicationFromConfigFile();
	void readPortsFromConfigFile();
	char * getUtilitySpecification(APP_TYPE appType);
	char * getUtilityConstant(APP_TYPE appType);
	
	void saveFile( string filename);
	void modifyFile(string filename);
	

	int sleepInterval;
	INT cTimer;
	
protected:
	template<class T> static string T_to_string( const T& t );
	template<class T> static T string_to_T( const string& s );
	
	static void trim( string& s );

public:

	struct key_not_found {  
		string key;
		key_not_found( const string& key_ = string() )
			: key(key_) {} };
};

unsigned int WINAPI startTimer(void *arg);
INT startTimerThread();

bool stopTimer(ConfigFile *configFile);
INT stopTimerThread();
INT initConfigFile();
INT reloadConfig();
bool isModified();

string getStringAppType(APP_TYPE appType);

template<class T>
string ConfigFile::T_to_string( const T& t )
{
	ostringstream oss;
	oss << t;
	return oss.str();
}

template<class T>
T ConfigFile::string_to_T( const string& s )
{
	T t;
	istringstream ist(s);
	ist >> t;
	if (s == ANY_CHARACTER) 
	{
		return ANY;
	}
	return t;
}

template<>
inline string ConfigFile::string_to_T<string>( const string& s )
{

	return s;
}

template<>
inline bool ConfigFile::string_to_T<bool>( const string& s )
{
	bool b = true;
	string sin = s;
	for( string::iterator p = sin.begin(); p != sin.end(); ++p )
		*p = toupper(*p); 
	if( sin==string("FALSE") || sin==string("0"))
		b = false;
	return b;
}


template<class T>
T ConfigFile::read( string key, string key2) 
{
	map_it p = myData.find(key);
	if( p == myData.end() ) {
		fprintf(stderr, "key not found\n");
		throw key_not_found(key);
	}
	
	map_it1d pp = myData[key].find(key2);
	if( p == myData.end() ) {
		fprintf(stderr, "key2 not found\n");
		throw key_not_found(key2);
	}
	if (myData[key][key2] == ANY_CHARACTER)
	{
		return string_to_T<T>(ANY_CHARACTER);
	}
	
	return string_to_T<T>( myData[key][key2]);
}

template<class T>
void ConfigFile::readFromKey( T& var, const string& key ) const
{
	map_it p = myData.find(key);
	if(p != myData.end()) var = string_to_T<T>( p->second );
}

template<class T>
void ConfigFile::add( string key, string key2, const T& value )
{
	string val = T_to_string( value );
	trim(key);
	trim(val);
	myData[key][key2] = val;
	return;
}

#endif 


