
#include "Capri.h"
#define MOVING_AVERAGE_SIZE 10
typedef enum 
{
	LAST_UTILITY = 1001,
	AVERAGE_MOVING = 1002,
	AVERAGE_INDEX = -1,
	AVERAGE_MEDION,
};



// initialize CRITICAL_SECTION
void initUtilityTable();


// end CRITICAL_SECTION
void closeUtilityTable();

// add utility tree to the utility table	
INT putUtilityTree(const INT fid, utilityTree *utilTree);

// add FlowID to the utility table
INT putFlowId(const INT fid, const FlowID *id);

// add Socket to the utility table
INT putSocket(const INT aid, const SOCKET newApp);

// add utility constant tree to the constraint table 
// utility constants include weights, min and max values
FLOAT putUtilityConstant(const INT fid, const utilityTree *c);

// get utility tree or utility constant tree from the table
utilityTree * getUtilityTree(const INT fid);

int setMax(const INT fid, int attr, float val);

int setCurMax(const INT fid, int attr, float val);

int setMin(const INT fid, int attr, float val);

// put attribute to the attribute table
INT putAttribute(const INT fid, INT attr, FLOAT val);

// put a normaized utility to the attribute table
INT putNormalizedUtility(const INT fid, float val);

// put utility specification string to the table (for default application)
INT putDefaultUtilitySpec(INT appType, const char *utilSpec, const char * utilConst);

// put constraint tree according to application type
INT putDefaultConstraintTree(APP_TYPE appType, utilityTree *constTree);

// put utility tree according to application type
INT putDefaultUtilityTree(APP_TYPE appType, utilityTree *utilTree);

// put port range 
INT putPortRange(APP_TYPE appType, INT start, INT end);

// get APP_TYPE
APP_TYPE getAppTypeByPortRange(INT port);

// get a FlowID from a given mapped index
const FlowID *getFlowId(const INT fid);

// get socket from a given mapped index
SOCKET getSocket(const INT aid);

// check if a mapped index (fid) is available
bool fidIsAvailable(const INT fid);

// get attribute from the attribute table
FLOAT getAttribute(const INT fid, INT attr);

// get utility value from the attribute table
FLOAT getUtility(const INT fid, INT utilIndex);

FLOAT getAveragedUtility(const INT fid, int action, int start_index, int window_size);

INT getUtilityIndex(int utilIndex, int shift);

FLOAT getUtilityDiff(const INT fid, int utilIndex, int shift);

// get a mapped index of a given FlowID
INT getTableIndex(const FlowID *id);

// get a mapped attribute index of a given unknown attribute string
INT getAttributeIndex(const INT fid, std::string attribute);

// get a utility speficiation
const char *getUtilitySpec(INT appType);

// get a utility constraint speficiation
const char *getConstraintSpec(INT appType);


// get a default utility tree by a given application type
utilityTree *getDefaultUtilityTree(APP_TYPE appType);

// get a default constraint tree by a given application type
utilityTree *getDefaultConstraintTree(APP_TYPE appType);

// get a free index from the utility table
INT	getFreeIndex();

// get a predefined mininum value of a given attribute 
FLOAT getMin(const INT fid, INT attr); 

// get a predefined maximum value of a given attribute 
FLOAT getMax(const INT fid, INT attr); 

FLOAT getCurMax(const INT fid, INT attr);

// check an attribute if available from the attribute table
bool checkAvailableConstraint(const INT fid, int attr);

//float getMinConstraint(const FlowID *id, int attr, int op); 
//float getMaxConstraint(const FlowID *id, int attr, int op); 

// deregister FlowID from the table
INT deregisterUtilityFromTable(const INT fid);

// get configurability: static or dynamic
INT getConfigurability(const INT fid);

// put configurability to the attribute table
INT putConfigurability(const INT fid, INT configurability);

// remove all statically registered configurations
INT removeStaticConfigurations(vector<FlowID> *removedFlowId);

// remove all registered ports
INT removePorts();

void printUtilityTable();
INT getUtilityTableSize();
mapUtility *getUtilityTable();
mapFlow *getFlowTable();
void printFlowID(const FlowID *Id);
void printFlowTable(INT i);
INT getConfigurabilityType(char *conf);

// lock utility table
void LockTable(); 

// release lock
void UnlockTable();  
